﻿using CommandLine;
using CommandLine.Text;
using NMF.Interop.Ecore;
using s = NMF.Interop.Ecore.Serialization;
using NMF.Interop.Ecore.Transformations;
using NMF.Transformations;
using NMF.Utilities;
using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Serialization;
using NMF.Transformations.Core;

namespace Ecore2Code
{
    class Options
    {
        public Options()
        {
            OverallNamespace = "Overall";
        }

        [Option('n', "namespace", HelpText="The parent namespace in case there are multiple root packages")]
        public string OverallNamespace { get; set; }

        [Option('l', "language", DefaultValue = SupportedLanguage.CS, HelpText = "The language in which the code should be generated")]
        public SupportedLanguage Language { get; set; }

        [ValueList(typeof(List<string>))]
        public IList<string> InputFiles { get; set; }

        [Option('o', "output", Required=true, HelpText="The output file in which the code should be generated")]
        public string OutputFile { get; set; }

        public string GetHelp()
        {
            return HelpText.AutoBuild(this);
        }
    }

    public enum SupportedLanguage
    {
        CS,
        VB,
        FS,
        CPP,
        JS
    }

    class Program
    {
        static void Main(string[] args)
        {
            Options options = new Options();
            if (Parser.Default.ParseArguments(args, options))
            {
#if DEBUG
                GenerateCode(options);
#else
                try
                {
                    GenerateCode(options);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
#endif
            }
            else
            {
                Console.WriteLine("Something went wrong!");
            }

            Console.ReadLine();
        }

        private static void GenerateCode(Options options)
        {
            var serializerTransform = new EcoreSerializerTransformation();
            var packageTransform = new EPackage2ClassesTransformation();

            var package = LoadPackageFromFiles(options.InputFiles, options.OverallNamespace);

            var transformedPackage = TransformationEngine.Transform<s.EPackage, EPackage>(package, serializerTransform);

            var compileUnit = TransformationEngine.Transform<EPackage, CodeCompileUnit>(transformedPackage, packageTransform);

            CodeDomProvider generator = null;

            switch (options.Language)
            {
                case SupportedLanguage.CS:
                    generator = new Microsoft.CSharp.CSharpCodeProvider();
                    break;
                case SupportedLanguage.VB:
                    generator = new Microsoft.VisualBasic.VBCodeProvider();
                    break;
                case SupportedLanguage.FS:
                    generator = new Microsoft.FSharp.Compiler.CodeDom.FSharpCodeProvider();
                    break;
                case SupportedLanguage.CPP:
                    generator = new Microsoft.VisualC.CppCodeProvider();
                    break;
                case SupportedLanguage.JS:
                    generator = new Microsoft.JScript.JScriptCodeProvider();
                    break;
                default:
                    generator = new Microsoft.CSharp.CSharpCodeProvider();
                    break;
            }

            using (var sw = new StreamWriter(options.OutputFile))
            {
                generator.GenerateCodeFromCompileUnit(compileUnit, sw, new System.CodeDom.Compiler.CodeGeneratorOptions()
                {
                    BlankLinesBetweenMembers = true,
                    VerbatimOrder = false,
                    ElseOnClosing = false
                });
            }

            Console.WriteLine("Code generated successfully!");
        }

        public static s.EPackage LoadPackageFromFiles(IList<string> files, string overallName)
        {
            var serializer = new XmlSerializer(typeof(NMF.Interop.Ecore.Serialization.EPackage));
            var serializer2 = new XmlSerializer(typeof(s.XMI), new Type[] { typeof(s.EPackage) });

            if (files == null || files.Count == 0) return null;

            var packages = new List<s.EPackage>();

            foreach (var ecoreFile in files)
            {
                try
                {
                    using (FileStream fs = new FileStream(ecoreFile, FileMode.Open, FileAccess.Read))
                    {
                        var package = serializer.Deserialize(fs) as s.EPackage;
                        if (package != null) packages.Add(package);
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("An error occurred reading the file. Continue to read as XMI. The error message was: " + ex.Message);
                    s.XMI xmi = null;

                    using (FileStream fs = new FileStream(files[0], FileMode.Open))
                    {
                        xmi = serializer2.Deserialize(fs) as s.XMI;
                    }

                    foreach (var item in xmi.Items)
                    {
                        var reader = new StringReader(item.OuterXml);
                        packages.Add(serializer.Deserialize(reader) as s.EPackage);
                    }
                }
            }

            if (packages.Count == 1)
            {
                return packages.First();
            }
            else
            {
                var package = new s.EPackage() { name = overallName };
                package.Items1 = packages.ToArray();
                return package;
            }
        }
    }
}
