﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using NMF.Transformations;
using NMF.Utilities;
using NMF.Serialization;

using System.CodeDom;
using System.ComponentModel;
using System.Collections.ObjectModel;
using NMF.Collections.ObjectModel;
using NMF.Transformations.Core;

namespace NMF.Interop.Ecore.Transformations
{
    public class EPackage2ClassesTransformation : ReflectiveTransformation
    {
        public class EPackage2CompileUnit : TransformationRule<EPackage, CodeCompileUnit>
        {
            public override CodeCompileUnit CreateOutput(EPackage input, ITransformationContext context)
            {
                return new CodeCompileUnit();
            }

            public override void Transform(EPackage input, CodeCompileUnit output, ITransformationContext context)
            {
                output.Namespaces.AddRange(context.Trace.ResolveInWhere(Rule<EPackage2Namespace>(), ns => ns.Classifiers.Count > 0).ToArray());
            }

            public override void RegisterDependencies()
            {
                Call(Rule<EPackage2Namespace>());
            }
        }

        public class EPackage2Namespace : TransformationRule<EPackage, CodeNamespace>
        {
            public override CodeNamespace CreateOutput(EPackage input, ITransformationContext context)
            {
                return new CodeNamespace(GetFullQualifiedName(input));
            }

            public override void Transform(EPackage input, CodeNamespace output, ITransformationContext context)
            {
                foreach (var cl in input.Classifiers.OfType<EClass>())
                {
                    var implementation = context.Trace.ResolveIn(Rule<EClass2ImplementationType>(), cl);

                    if (!output.Types.Contains(implementation))
                    {
                        output.Types.Add(implementation);
                    }
                }

                var referenceTypes = context.Trace.ResolveInWhere(Rule<EReference2Type>(), r => r.DeclaringClass.Package == input);

                foreach (var type in referenceTypes)
                {
                    output.Types.Add(type);
                }

                output.Imports.Add(new CodeNamespaceImport("NMF.Collections.ObjectModel"));
                output.Imports.Add(new CodeNamespaceImport("System.Collections.ObjectModel"));

                var otherNamespaces = context.Trace.ResolveInWhere(Rule<EPackage2Namespace>(), ns => ns.Classifiers.Count > 0 && ns != input)
                    .Select(ns => new CodeNamespaceImport(ns.Name));

                output.Imports.AddRange(otherNamespaces.ToArray());
            }

            private string GetFullQualifiedName(EPackage input)
            {
                var namesp = input.Name.ToPascalCase();
                if (input.ParentPackage != null)
                {
                    return GetFullQualifiedName(input.ParentPackage) + "." + namesp;
                }
                else
                {
                    return namesp;
                }
            }

            public override void RegisterDependencies()
            {
                RequireMany(Rule<EPackage2Namespace>(),
                    package => package.SubPackages);

                TransformationDelayLevel = 1;
            }
        }

        public class EClassifier2Type : AbstractTransformationRule<EClassifier, CodeTypeDeclaration>
        {
            public override void Transform(EClassifier input, CodeTypeDeclaration output, ITransformationContext context)
            {
                output.Attributes = MemberAttributes.Public;
            }

            public override void RegisterDependencies()
            {
                CallForEach<EPackage, CodeNamespace>(
                    package => package.Classifiers,
                    (ns, types) => {
                        foreach (var type in types)
                        {
                            ns.Types.Add(type);
                        }
                    });
            }
        }

        public class EClass2Type : TransformationRule<EClass, CodeTypeDeclaration>
        {
            public override CodeTypeDeclaration CreateOutput(EClass input, ITransformationContext context)
            {
                return new CodeTypeDeclaration()
                {
                    Name = input.Name.ToPascalCase(),
                    IsClass = !input.IsInterface,
                    IsInterface = input.IsInterface
                };
            }

            public override void Transform(EClass input, CodeTypeDeclaration output, ITransformationContext context)
            {
                int baseClasses = input.SuperTypes.Count(cl => !cl.IsInterface);
                if (baseClasses > 1)
                {
                    output.IsInterface = true;

                    EnsureTypesAreInterfaces(input, context, type => type.SuperTypes);
                }

                foreach (var baseType in input.SuperTypes)
                {
                    var decl = context.Trace.ResolveIn(Rule<EClass2Type>(), baseType);
                    output.BaseTypes.Add(decl.Name);
                }

                foreach (var derived in input.DerivedTypes)
                {
                    var decl = context.Trace.ResolveIn(Rule<EClass2Type>(), derived);
                    output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(XmlKnownTypeAttribute)),
                        new CodeAttributeArgument(new CodeTypeOfExpression(derived.Name.ToPascalCase()))));
                }

                context.CallTransformation(Rule<EClass2ImplementationType>(), input);

                output.IsEnum = false;
                output.IsPartial = true;
                output.IsStruct = false;
                
                if (input.IsAbstract)
                {
                    output.Attributes |= MemberAttributes.Abstract;
                }

                AddAttribute(output, typeof(XmlElementNameAttribute), input.Name);
                AddAttribute(output, typeof(XmlNamespaceAttribute), input.Package.NsUri);
                AddAttribute(output, typeof(XmlNamespacePrefixAttribute), input.Package.NsPrefix);
            }

            private void EnsureTypesAreInterfaces(EClass type, ITransformationContext context, Func<EClass, IEnumerable<EClass>> children)
            {
                foreach (var childType in children(type))
                {
                    var transformed = context.Trace.ResolveIn(Rule<EClass2Type>(), childType);
                    if (!transformed.IsInterface)
                    {
                        transformed.IsInterface = true;
                        EnsureTypesAreInterfaces(childType, context, children);
                    }
                }
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EClassifier2Type>());

                RequireMany(Rule<EClass2Type>(), cl => cl.DerivedTypes);

                CallMany(Rule<EOperation2Method>(), cl => cl.Operations);

                CallMany(Rule<EStructuralFeature2Property>(), cl => cl.Features);
            }
        }

        public class EClass2ImplementationType : TransformationRule<EClass, CodeTypeDeclaration>
        {
            public override CodeTypeDeclaration CreateOutput(EClass input, ITransformationContext context)
            {
                var def = context.Trace.ResolveIn(Rule<EClass2Type>(), input);
                if (def == null || !def.IsInterface)
                {
                    return def;
                }
                else
                {
                    var decl = new CodeTypeDeclaration(input.Name.ToPascalCase() + "Implementation");
                    decl.IsClass = true;
                    decl.BaseTypes.Add(def.Name);
                    def.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(XmlDefaultImplementationTypeAttribute)), 
                        new CodeAttributeArgument(new CodeTypeOfExpression(decl.Name))));
                    return decl;
                }
            }

            public override void RegisterDependencies()
            {
                TransformationDelayLevel = 2;

                Require(Rule<EClass2Constructor>(), (decl, con) => decl.Members.Add(con));
            }

            public override void Transform(EClass input, CodeTypeDeclaration output, ITransformationContext context)
            {
                var original = context.Trace.ResolveIn(Rule<EClass2Type>(), input);

                var constructor = context.Trace.ResolveIn(Rule<EClass2Constructor>(), input);

                AddBaseClassImplementations(input, output, output == original, constructor, context);
            }

            private void AddBaseClassImplementations(EClass cl, CodeTypeDeclaration output, bool isOriginal, CodeConstructor constructor, ITransformationContext context)
            {
                foreach (var baseType in cl.SuperTypes)
                {
                    var original = context.Trace.ResolveIn(Rule<EClass2Type>(), baseType);
                    var implementation = context.Trace.ResolveIn(Rule<EClass2ImplementationType>(), baseType);

                    if (original != implementation)
                    {
                        foreach (CodeTypeMember member in implementation.Members)
                        {
                            if (!(member is CodeConstructor))
                            {
                                if (!output.Members.Contains(member)) output.Members.Add(member);
                            }
                            else
                            {
                                var con = member as CodeConstructor;
                                foreach (CodeStatement stmt in con.Statements)
                                {
                                    constructor.Statements.Add(stmt);
                                }
                            }
                        }
                        AddBaseClassImplementations(baseType, output, false, constructor, context);
                    }
                    else
                    {
                        if (!isOriginal) output.BaseTypes.Add(implementation.Name);
                    }
                }
            }
        }

        public class EEnum2Type : TransformationRule<EEnum, CodeTypeDeclaration>
        {
            public override CodeTypeDeclaration CreateOutput(EEnum input, ITransformationContext context)
            {
                return new CodeTypeDeclaration()
                {
                    Name = input.Name.ToPascalCase(),
                    IsEnum = true
                };
            }

            public override void Transform(EEnum input, CodeTypeDeclaration output, ITransformationContext context)
            {
                output.IsEnum = true;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EDataType2Type>());

                RequireMany(Rule<EEnumLiteral2Field>(), en => en.Literals,
                    (decl, literals) =>
                    {
                        foreach (var literal in literals)
                        {
                            decl.Members.Add(literal);
                        }
                    });
            }
        }

        public class EDataType2Type : TransformationRule<EDataType, CodeTypeDeclaration>
        {
            public override CodeTypeDeclaration CreateOutput(EDataType input, ITransformationContext context)
            {
                return new CodeTypeDeclaration()
                {
                    Name = input.Name.ToPascalCase(),
                    IsStruct = true
                };
            }

            public override void Transform(EDataType input, CodeTypeDeclaration output, ITransformationContext context)
            {
                output.IsStruct = true;
                output.IsEnum = false;
                output.IsPartial = true;
                output.IsClass = false;
                output.IsInterface = false;

                if (input.IsSerializable)
                {
                    output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(SerializableAttribute))));
                }
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EClassifier2Type>());
            }
        }

        public class EEnumLiteral2Field : TransformationRule<EEnumLiteral, CodeMemberField>
        {
            public override void Transform(EEnumLiteral input, CodeMemberField output, ITransformationContext context)
            {
                output.Name = input.Name.ToPascalCase();
                output.InitExpression = new CodePrimitiveExpression(input.Value);
            }
        }

        private static CodeTypeReference CreateTypeReference(ETypedElement typedElement, Action<CodeAttributeDeclaration> attributePersistor, Func<ETypedElement, ITransformationContext, CodeTypeReference> getCollectionType, ITransformationContext context)
        {
            if (typedElement == null) return new CodeTypeReference(typeof(void));

            if (typedElement.Type is MappableDataType)
            {
                if (typedElement.UpperBound == 1)
                {
                    return new CodeTypeReference((typedElement.Type as MappableDataType).Type);
                }
                else
                {
                    return new CodeTypeReference(typeof(ObservableCollection<>).MakeGenericType(new Type[] { (typedElement.Type as MappableDataType).Type }));
                }
            }
            else
            {
                var classifier = context.Trace.ResolveIn((EClassifier2Type)context.Transformation.GetRuleForRuleType(typeof(EClassifier2Type)), typedElement.Type);
                if (typedElement.UpperBound == 1)
                {
                    if (classifier == null) return new CodeTypeReference(typeof(object));
                    return new CodeTypeReference(classifier.Name);
                }
                else
                {
                    if (typedElement.LowerBound != 0) attributePersistor(new CodeAttributeDeclaration(new CodeTypeReference(typeof(LowerBoundAttribute)), new CodeAttributeArgument(new CodePrimitiveExpression(typedElement.UpperBound))));
                    if (typedElement.UpperBound != -1) attributePersistor(new CodeAttributeDeclaration(new CodeTypeReference(typeof(UpperBoundAttribute)), new CodeAttributeArgument(new CodePrimitiveExpression(typedElement.UpperBound))));
                    if (getCollectionType == null)
                    {
                        if (classifier == null) return new CodeTypeReference(typeof(IEnumerable<object>));
                        return new CodeTypeReference(typeof(IEnumerable<>).Name, new CodeTypeReference(classifier.Name));
                    }
                    else
                    {
                        return getCollectionType(typedElement, context);
                    }
                }
            }
        }

        public class EOperation2Method : TransformationRule<EOperation, CodeMemberMethod>
        {
            public override void Transform(EOperation input, CodeMemberMethod output, ITransformationContext context)
            {
                output.Name = input.Name.ToPascalCase();

                if (input.Type != null)
                {
                    output.ReturnType = CreateTypeReference(input, attr => output.CustomAttributes.Add(attr), null, context);
                    output.Statements.Add(new CodeMethodReturnStatement(new CodeDefaultValueExpression(output.ReturnType)));
                }
                else
                {
                    output.ReturnType = new CodeTypeReference(typeof(void));
                }


                var classDeclaration = context.Trace.ResolveIn(Rule<EClass2Type>(), input.DeclaringClass);
                var implDeclaration = context.Trace.ResolveIn(Rule<EClass2ImplementationType>(), input.DeclaringClass);

                if (classDeclaration != implDeclaration)
                {
                    var clone = new CodeMemberMethod()
                    {
                        Name = output.Name,
                        ReturnType = output.ReturnType
                    };
                    foreach (CodeParameterDeclarationExpression par in output.Parameters)
                    {
                        clone.Parameters.Add(par);
                    }
                    classDeclaration.Members.Add(clone);
                    output.PrivateImplementationType = new CodeTypeReference(classDeclaration.Name);
                }
                else
                {
                    output.Attributes = MemberAttributes.Public;
                }
                implDeclaration.Members.Add(output);
            }

            public override void RegisterDependencies()
            {
                RequireMany(Rule<EParameter2Parameter>(), op => op.Parameters,
                    (op, parameters) =>
                    {
                        foreach (var par in parameters)
                        {
                            op.Parameters.Add(par);
                        }
                    });

                TransformationDelayLevel = 1;
            }
        }

        public class EParameter2Parameter : TransformationRule<EParameter, CodeParameterDeclarationExpression>
        {
            public override void Transform(EParameter input, CodeParameterDeclarationExpression output, ITransformationContext context)
            {
                output.Direction = FieldDirection.In;
                output.Name = input.Name.ToCamelCase();

                output.Type = CreateTypeReference(input, attr => output.CustomAttributes.Add(attr), null, context);
            }

            public override void RegisterDependencies()
            {
                base.RegisterDependencies();
            }
        }

        private static void AddAttribute(CodeTypeMember declaration, Type attributeType, object argument)
        {
            declaration.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(attributeType), new CodeAttributeArgument(new CodePrimitiveExpression(argument))));
        }

        public class EStructuralFeature2Property : TransformationRule<EStructuralFeature, CodeMemberProperty>
        {
            public override void Transform(EStructuralFeature input, CodeMemberProperty output, ITransformationContext context)
            {
                output.Attributes = MemberAttributes.Public;
                output.Name = input.Name.ToPascalCase();

                var field = CreateField(input, output.Type);
                var fieldRef = new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), field.Name);
                var eventField = CreateEvent(input);
                var eventMethod = CreateEventMethod(input);

                output.HasGet = true;

                var constructor = context.Trace.ResolveIn(Rule<EClass2Constructor>(), input.DeclaringClass);

                output.GetStatements.Add(new CodeMethodReturnStatement(fieldRef));

                var classDeclaration = context.Trace.ResolveIn(Rule<EClass2Type>(), input.DeclaringClass);
                var implDeclaration = context.Trace.ResolveIn(Rule<EClass2ImplementationType>(), input.DeclaringClass);

                if (input.UpperBound == 1)
                {
                    CreateSetStatement(input, output, fieldRef);

                    implDeclaration.Members.Add(eventField);
                    implDeclaration.Members.Add(eventMethod);

                    if (classDeclaration != implDeclaration)
                    {
                        var newEvent = CreateEvent(input);
                        newEvent.Attributes = MemberAttributes.Final;
                        classDeclaration.Members.Add(newEvent);
                    }
                }
                else
                {
                    if (input is EReference && (input as EReference).Opposite != null)
                    {
                        constructor.Statements.Add(new CodeAssignStatement(fieldRef, new CodeObjectCreateExpression(output.Type, new CodeThisReferenceExpression())));
                    }
                    else
                    {
                        constructor.Statements.Add(new CodeAssignStatement(fieldRef, new CodeObjectCreateExpression(output.Type)));
                    }

                    output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(DesignerSerializationVisibilityAttribute)),
                        new CodeAttributeArgument(new CodeFieldReferenceExpression(new CodeTypeReferenceExpression(typeof(DesignerSerializationVisibility)), DesignerSerializationVisibility.Content.ToString()))));
                }

                
                implDeclaration.Members.Add(field);
                classDeclaration.Members.Add(output);
                if (implDeclaration != classDeclaration)
                {
                    implDeclaration.Members.Add(output);
                }

                AddAttribute(output, typeof(XmlElementNameAttribute), input.Name);
                AddAttribute(output, typeof(XmlNamespaceAttribute), null);
                AddAttribute(output, typeof(XmlNamespacePrefixAttribute), null);
            }

            private void CreateSetStatement(EStructuralFeature input, CodeMemberProperty output, CodeExpression field)
            {
                var ifStmt = new CodeConditionStatement();
                var val = new CodePropertySetValueReferenceExpression();
                ifStmt.Condition = new CodeBinaryOperatorExpression(field, CodeBinaryOperatorType.IdentityInequality, val);

                var assignment = new CodeAssignStatement(field, val);
                

                if (input is EReference)
                {
                    var reference = input as EReference;

                    if (reference.Opposite != null)
                    {
                        ifStmt.TrueStatements.Add(new CodeVariableDeclarationStatement(output.Type, "old", field));
                        var old = new CodeVariableReferenceExpression("old");
                        var nullRef = new CodePrimitiveExpression(null);
                        var oldNotNull = new CodeBinaryOperatorExpression(old, CodeBinaryOperatorType.IdentityInequality, nullRef);
                        var valueNotNull = new CodeBinaryOperatorExpression(val, CodeBinaryOperatorType.IdentityInequality, nullRef);
                        if (reference.Opposite.UpperBound == 1)
                        {
                            ifStmt.TrueStatements.Add(assignment);

                            var oldCheck = new CodeConditionStatement(oldNotNull, new CodeAssignStatement
                                (new CodePropertyReferenceExpression(old, reference.Opposite.Name), nullRef));

                            var newCheck = new CodeConditionStatement(valueNotNull, new CodeAssignStatement
                                (new CodePropertyReferenceExpression(val, reference.Opposite.Name), val));

                            ifStmt.TrueStatements.Add(oldCheck);
                            ifStmt.TrueStatements.Add(newCheck);
                        }
                        else
                        {
                            ifStmt.TrueStatements.Add(assignment);

                            var oldCheck = new CodeConditionStatement(oldNotNull, new CodeExpressionStatement(new CodeMethodInvokeExpression
                                (new CodePropertyReferenceExpression(old, reference.Opposite.Name.ToPascalCase()), "Remove", new CodeThisReferenceExpression())));

                            var newCheck = new CodeConditionStatement(valueNotNull, new CodeExpressionStatement(new CodeMethodInvokeExpression
                                (new CodePropertyReferenceExpression(val, reference.Opposite.Name.ToPascalCase()), "Add", new CodeThisReferenceExpression())));

                            ifStmt.TrueStatements.Add(oldCheck);
                            ifStmt.TrueStatements.Add(newCheck);
                        }
                    }
                    else
                    {
                        ifStmt.TrueStatements.Add(assignment);
                    }
                }
                else
                {
                    ifStmt.TrueStatements.Add(assignment);
                }

                ifStmt.TrueStatements.Add(new CodeMethodInvokeExpression(new CodeThisReferenceExpression(), "On" + output.Name.ToPascalCase() + "Changed", new CodeObjectCreateExpression(typeof(EventArgs))));
                output.SetStatements.Add(ifStmt);
                output.HasSet = true;
            }

            private CodeMemberField CreateField(EStructuralFeature input, CodeTypeReference codeTypeReference)
            {
                return new CodeMemberField(codeTypeReference, input.Name.ToCamelCase());
            }

            private static CodeMemberEvent CreateEvent(EStructuralFeature input)
            {
                var ev = new CodeMemberEvent();

                ev.Attributes = MemberAttributes.Public | MemberAttributes.Final;
                ev.Name = input.Name.ToPascalCase() + "Changed";
                ev.Type = new CodeTypeReference(typeof(EventHandler));

                return ev;
            }

            private static CodeMemberMethod CreateEventMethod(EStructuralFeature input)
            {
                var eventMethod = new CodeMemberMethod();

                eventMethod.Attributes = MemberAttributes.Family;

                eventMethod.Name = "On" + input.Name.ToPascalCase() + "Changed";

                eventMethod.Parameters.Add(new CodeParameterDeclarationExpression(typeof(EventArgs), "eventArgs"));

                var ifStmt = new CodeConditionStatement();
                var eventRef = new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), input.Name.ToPascalCase() + "Changed");
                ifStmt.Condition = new CodeBinaryOperatorExpression(eventRef, CodeBinaryOperatorType.IdentityInequality, new CodePrimitiveExpression(null));
                ifStmt.TrueStatements.Add(new CodeMethodInvokeExpression(eventRef, "Invoke", new CodeThisReferenceExpression(), new CodeArgumentReferenceExpression("eventArgs")));

                eventMethod.Statements.Add(ifStmt);

                return eventMethod;
            }

            public override void RegisterDependencies()
            {
                TransformationDelayLevel = 1;
            }
        }

        public class EClass2Constructor : TransformationRule<EClass, CodeConstructor>
        {
            public override void Transform(EClass input, CodeConstructor output, ITransformationContext context)
            {
                output.Attributes = MemberAttributes.Public;
            }
        }

        public class EReference2Property : TransformationRule<EReference, CodeMemberProperty>
        {
            public override CodeMemberProperty CreateOutput(EReference input, ITransformationContext context)
            {
                var prop = new CodeMemberProperty();

                prop.Type = CreateTypeReference(input, attr => prop.CustomAttributes.Add(attr), CreateCollectionType, context);

                return prop;
            }

            public override void Transform(EReference input, CodeMemberProperty output, ITransformationContext context)
            {
                if (input.IsContainment)
                {
                    output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(ContainmentAttribute))));
                    AddAttribute(output, typeof(XmlAttributeAttribute), false);
                }
                else
                {
                    if (input.Opposite != null && input.Opposite.IsContainment)
                    {
                        if (input.UpperBound == 1)
                        {
                            output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(DesignerSerializationVisibilityAttribute)), new CodeAttributeArgument(
                                new CodeFieldReferenceExpression(new CodeTypeReferenceExpression(typeof(DesignerSerializationVisibility)), "Hidden"))));
                        }
                        else
                        {
                            var attDecl = output.CustomAttributes.OfType<CodeAttributeDeclaration>().FirstOrDefault(att => att.AttributeType.BaseType == typeof(DesignerSerializationVisibilityAttribute).Name);
                            if (attDecl == null)
                            {
                                output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(DesignerSerializationVisibilityAttribute)), new CodeAttributeArgument(
                                    new CodeFieldReferenceExpression(new CodeTypeReferenceExpression(typeof(DesignerSerializationVisibility)), "Hidden"))));
                            }
                            else
                            {
                                attDecl.Arguments[0].Value = new CodeFieldReferenceExpression(new CodeTypeReferenceExpression(typeof(DesignerSerializationVisibility)), "Hidden");
                            }
                        }
                    }
                    AddAttribute(output, typeof(XmlAttributeAttribute), true);
                }
            }

            private CodeTypeReference CreateCollectionType(ETypedElement arg, ITransformationContext context)
            {
                var reference = arg as EReference;

                if (reference.Opposite != null)
                {
                    return new CodeTypeReference(
                        context.Trace.ResolveIn(Rule<EReference2Type>(), reference).Name);
                }
                else
                {
                    if (reference.Type != null)
                    {
                        return new CodeTypeReference(typeof(ObservableCollection<>).Name, new CodeTypeReference(reference.Type.Name.ToPascalCase()));
                    }
                    else
                    {
                        return new CodeTypeReference(typeof(ObservableCollection<object>));
                    }
                }
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EStructuralFeature2Property>());

                OutputDelayLevel = 1;

                Require(Rule<EReference2Type>(), reference => reference.UpperBound != 1 && reference.Opposite != null);
            }
        }

        public class EAttribute2Property : TransformationRule<EAttribute, CodeMemberProperty>
        {
            public override CodeMemberProperty CreateOutput(EAttribute input, ITransformationContext context)
            {
                var prop = new CodeMemberProperty();
                prop.Type = CreateTypeReference(input, attr => prop.CustomAttributes.Add(attr), CreateCollectionType, context);
                return prop;
            }

            public override void Transform(EAttribute input, CodeMemberProperty output, ITransformationContext context)
            {
                if (input.IsId)
                {
                    output.CustomAttributes.Add(new CodeAttributeDeclaration(new CodeTypeReference(typeof(IdAttribute))));
                    var declaration = context.Trace.ResolveIn(Rule<EClass2Type>(), input.DeclaringClass);
                    AddAttribute(declaration, typeof(XmlIdentifierAttribute), input.Name);
                }
                AddAttribute(output, typeof(XmlAttributeAttribute), true);
            }

            private CodeTypeReference CreateCollectionType(ETypedElement arg, ITransformationContext context)
            {
                return new CodeTypeReference(typeof(ObservableCollection<>).Name, new CodeTypeReference(arg.Type.Name.ToPascalCase()));
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EStructuralFeature2Property>());
            }
        }

        public class EReference2Type : TransformationRule<EReference, CodeTypeDeclaration>
        {
            public override CodeTypeDeclaration CreateOutput(EReference input, ITransformationContext context)
            {
                return new CodeTypeDeclaration()
                {
                    Name = input.DeclaringClass.Name.ToPascalCase() + input.Name.ToPascalCase() + "Collection"
                };
            }

            public override void Transform(EReference input, CodeTypeDeclaration output, ITransformationContext context)
            {
                output.IsClass = true;
                output.Attributes = MemberAttributes.Public;

                Type collectionTypeTemplate;

                if (input.IsOrdered)
                {
                    if (input.IsUnique)
                    {
                        collectionTypeTemplate = typeof(ObservableOppositeOrderedSet<,>);
                    }
                    else
                    {
                        collectionTypeTemplate = typeof(ObservableOppositeList<,>);
                    }
                }
                else
                {
                    collectionTypeTemplate = typeof(ObservableOppositeSet<,>);
                }

                output.BaseTypes.Add(new CodeTypeReference(collectionTypeTemplate.FullName, new CodeTypeReference(input.DeclaringClass.Name.ToPascalCase()), new CodeTypeReference(input.Type.Name.ToPascalCase())));

                var typeRef = new CodeTypeReference(input.DeclaringClass.Name.ToPascalCase());
                var collRef = new CodeTypeReference(input.Type.Name.ToPascalCase());

                var constructor = new CodeConstructor();
                constructor.Attributes = MemberAttributes.Public;
                constructor.Parameters.Add(new CodeParameterDeclarationExpression(typeRef, "parent"));
                constructor.BaseConstructorArgs.Add(new CodeArgumentReferenceExpression("parent"));
                output.Members.Add(constructor);

                var method = new CodeMemberMethod();
                method.Attributes = MemberAttributes.Family | MemberAttributes.Override;
                method.Name = "SetOpposite";
                method.Parameters.Add(new CodeParameterDeclarationExpression(collRef, "item"));
                method.Parameters.Add(new CodeParameterDeclarationExpression(typeRef, "parent"));

                var opposite = input.Opposite;

                var item = new CodeArgumentReferenceExpression("item");
                var parent = new CodeArgumentReferenceExpression("parent");
                
                if (opposite.UpperBound == 1)
                {
                    method.Statements.Add(new CodeAssignStatement(new CodePropertyReferenceExpression(item, opposite.Name.ToPascalCase()), parent));
                }
                else
                {
                    var ifStmt = new CodeConditionStatement();
                    ifStmt.Condition = new CodeBinaryOperatorExpression(parent, CodeBinaryOperatorType.IdentityEquality, new CodePrimitiveExpression(null));
                    var item_opp = new CodePropertyReferenceExpression(item, opposite.Name.ToPascalCase());
                    ifStmt.TrueStatements.Add(new CodeMethodInvokeExpression(item_opp, "Remove", new CodePropertyReferenceExpression(new CodeThisReferenceExpression(), "Parent")));
                    ifStmt.FalseStatements.Add(new CodeMethodInvokeExpression(item_opp, "Add", parent));
                    method.Statements.Add(ifStmt);
                }
                output.Members.Add(method);
            }
        }


    }
}
