﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NMF.Transformations;
using NMF.Utilities;
using NMF.Transformations.Core;

namespace NMF.Interop.Ecore.Transformations
{
    public class EcoreSerializerTransformation : ReflectiveTransformation
    {
        private Dictionary<string, EClassifier> standardTypes = new Dictionary<string, EClassifier>();

        protected virtual void CreateStandardTypes(IDictionary<string, EClassifier> standardTypes)
        {
            standardTypes.Add("EString", DefaultTypes.EString);
            standardTypes.Add("EBigDecimal", new EClass()
            {
                Name = "EBigDecimal"
            });
            standardTypes.Add("EBigInteger", DefaultTypes.EBigInteger);
            standardTypes.Add("EBoolean", DefaultTypes.EBoolean);
            standardTypes.Add("EBooleanObject", DefaultTypes.EBoolean);
            standardTypes.Add("EByte", DefaultTypes.EByte);
            standardTypes.Add("EByteArray", DefaultTypes.EByteArray);
            standardTypes.Add("EByteObject", new EClass()
            {
                Name = "EByteObject"
            });
            standardTypes.Add("EChar", DefaultTypes.EChar);
            standardTypes.Add("ECharacterObject", DefaultTypes.EChar);
            standardTypes.Add("EClass", DefaultTypes.EClass);
            standardTypes.Add("EDate", DefaultTypes.EDate);
            standardTypes.Add("EDiagnosticChain", new EClass()
            {
                Name = "EDiagnosticChain"
            });
            standardTypes.Add("EDouble", DefaultTypes.EDouble);
            standardTypes.Add("EDoubleObject", DefaultTypes.EDouble);
            standardTypes.Add("EEList<E>", new EClass()
            {
                Name = "EEList<E>",
            });
            standardTypes.Add("EEnumerator", new EClass()
            {
                Name = "EEnumerator"
            });
            standardTypes.Add("EFeatureMap", new EClass()
            {
                Name = "EFeatureMap"
            });
            standardTypes.Add("EFeatureMapEntry", new EClass()
            {
                Name = "EFeatureMapEntry"
            });
            standardTypes.Add("EFloat", DefaultTypes.EFloat);
            standardTypes.Add("EFloatObject", DefaultTypes.EFloat);
            standardTypes.Add("EInt", DefaultTypes.EInt);
            standardTypes.Add("EIntegerObject", DefaultTypes.EInt);
            standardTypes.Add("EJavaClass<T>", new EClass()
            {
                Name = "EJavaClass<T>"
            });
            standardTypes.Add("EJavaObject", new EClass()
            {
                Name = "EJavaObject"
            });
            standardTypes.Add("ELong", DefaultTypes.ELong);
            standardTypes.Add("ELongObject", DefaultTypes.ELong);
            standardTypes.Add("EMap<K,V>", new EClass()
            {
                Name = "EMap<K,V>"
            });
            standardTypes.Add("EObject", DefaultTypes.EObject);
            standardTypes.Add("EResource", new EClass()
            {
                Name = "EResource"
            });
            standardTypes.Add("EResourceSet", new EClass()
            {
                Name = "EResourceSet"
            });
            standardTypes.Add("EShort", DefaultTypes.EShort);
            standardTypes.Add("EShortObject", DefaultTypes.EShort);
            standardTypes.Add("ETreeIterator<E>", new EClass()
            {
                Name = "ETreeIterator<E>"
            });
            standardTypes.Add("EInvocationTargetException", new EClass()
            {
                Name = "EInvocationTargetException"
            });

            var package = new EPackage()
            {
                Name = "Ecore",
                NsPrefix = "ecore",
                NsUri = "http://www.eclipse.org/emf/2002/Ecore",
                Id = "Ecore"
            };

            foreach (var type in standardTypes.Values)
            {
                package.Classifiers.Add(type);
            }
        }

        public EcoreSerializerTransformation()
        {
            CreateStandardTypes(standardTypes);
        }

        public virtual EClass TraceClass(string path, ITransformationContext context)
        {
            EClass result = null;
            string[] segments = path.Trim().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
            if (segments != null && segments.Length >= 1)
            {
                var className = segments[segments.Length - 1];

                result = context.Trace.FindInWhere((EClassTransformation)GetRuleForRuleType(typeof(EClassTransformation)),
                    c => c.Name == className)
                    .FirstOrDefault();
            }
            var e = new ResolveClassEventArgs(path, context) { Class = result };
            OnResolveClass(e);
            return e.Class;
        }

        public EClassifier TraceType(string path, ITransformationContext context)
        {
            EClassifier result = null;
            string[] segments = path.Trim().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
            if (segments != null && segments.Length >= 1)
            {
                var className = segments[segments.Length - 1];

                if (standardTypes.ContainsKey(className))
                {
                    result = standardTypes[className];
                }
                else
                {
                    result = context.Trace.FindInWhere((EClassifierTransformation)GetRuleForRuleType(typeof(EClassifierTransformation)),
                        c => c.Name == className).FirstOrDefault();
                }
            }
            var e = new ResolveTypeEventArgs(path, context) { Type = result };
            OnResolveType(e);
            return e.Type;
        }

        public virtual EReference TraceReference(string path, ITransformationContext context)
        {
            EReference result = null;
            string[] segments = path.Trim().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
            if (segments != null && segments.Length >= 2)
            {
                var referenceName = segments[segments.Length - 1];
                var className = segments[segments.Length - 2];

                result = context.Trace.FindInWhere((EReferenceTransformation)GetRuleForRuleType(typeof(EReferenceTransformation)),
                    r => r.Name == referenceName && r.DeclaringClass != null && r.DeclaringClass.Name == className).FirstOrDefault();
            }
            var e = new ResolveReferenceEventArgs(path, context) { Reference = result };
            OnResolveReference(e);
            return e.Reference;
        }

        protected virtual void OnResolveClass(ResolveClassEventArgs e)
        {
            if (ResolveClass != null) ResolveClass(this, e);
        }

        protected virtual void OnResolveType(ResolveTypeEventArgs e)
        {
            if (ResolveType != null) ResolveType(this, e);
        }

        protected virtual void OnResolveReference(ResolveReferenceEventArgs e)
        {
            if (ResolveReference != null) ResolveReference(this, e);
        }

        public event EventHandler<ResolveClassEventArgs> ResolveClass;
        public event EventHandler<ResolveTypeEventArgs> ResolveType;
        public event EventHandler<ResolveReferenceEventArgs> ResolveReference;


        public class EAnnotationTransformation : TransformationRule<Serialization.EAnnotation, EAnnotation>
        {

            public override void Transform(Serialization.EAnnotation input, EAnnotation output, ITransformationContext context)
            {
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EModelElementTransformation>());

                CallForEach<Serialization.EModelElement, EModelElement>(
                    me => me.Items, (me, annotations) => me.Annotations.AddRange(annotations));
            }
        }

        public class EAttributeTransformation : TransformationRule<Serialization.EAttribute, EAttribute>
        {

            public override void Transform(Serialization.EAttribute input, EAttribute output, ITransformationContext context)
            {
                output.IsId = input.iD;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EStructuralFeatureTransformation>());
            }
        }

        public class EClassTransformation : TransformationRule<Serialization.EClass, EClass>
        {

            public override void Transform(Serialization.EClass input, EClass output, ITransformationContext context)
            {
                output.IsAbstract = input.@abstract;
                output.IsInterface = input.interfaceSpecified;

                if (input.eSuperTypes != null)
                {
                    var types = input.eSuperTypes.Split(' ');
                    foreach (var type in types)
                    {
                        var eClass = (Transformation as EcoreSerializerTransformation).TraceClass(type, context);
                        if (eClass != null) output.SuperTypes.Add(eClass);
                    }
                }
            }

            public override EClass CreateOutput(Serialization.EClass input, ITransformationContext context)
            {
                return new EClass() { Name = input.name };
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EClassifierTransformation>());
            }
        }

        public class EClassifierTransformation : AbstractTransformationRule<Serialization.EClassifier, EClassifier>
        {
            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ENamedElementTransformation>());

                CallForEach<Serialization.EPackage, EPackage>(
                    pck => pck.Items1.AllOfType<Serialization.EClassifier>(), 
                    (pck, classes) => pck.Classifiers.AddRange(classes));
            }

            public override void Transform(Serialization.EClassifier input, EClassifier output, ITransformationContext context)
            {
                output.InstanceClassName = input.instanceClassName;
            }
        }

        public class EDataTypeTransformation : TransformationRule<Serialization.EDataType, EDataType>
        {
            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EClassifierTransformation>());
            }

            public override EDataType CreateOutput(Serialization.EDataType input, ITransformationContext context)
            {
                return new EDataType() { Name = input.name };
            }

            public override void Transform(Serialization.EDataType input, EDataType output, ITransformationContext context)
            {
                output.IsSerializable = input.serializable;
            }
        }

        public class EEnumTransformation : TransformationRule<Serialization.EEnum, EEnum>
        {
            public override EEnum CreateOutput(Serialization.EEnum input, ITransformationContext context)
            {
                return new EEnum() { Name = input.name };
            }

            public override void Transform(Serialization.EEnum input, EEnum output, ITransformationContext context)
            {
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EDataTypeTransformation>());
            }
        }

        public class EEnumLiteralTransformation : TransformationRule<Serialization.EEnumLiteral, EEnumLiteral>
        {
            public override void Transform(Serialization.EEnumLiteral input, EEnumLiteral output, ITransformationContext context)
            {
                output.Value = input.value;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ENamedElementTransformation>());

                CallForEach<Serialization.EEnum, EEnum>(
                    en => en.Items1, (en, literals) => en.Literals.AddRange(literals));
            }
        }

        public class EFactoryTransformation : TransformationRule<Serialization.EFactory, EFactory>
        {
            public override void Transform(Serialization.EFactory input, EFactory output, ITransformationContext context)
            {
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EModelElementTransformation>());
            }
        }

        public class EModelElementTransformation : AbstractTransformationRule<Serialization.EModelElement, EModelElement>
        {
            public override void Transform(Serialization.EModelElement input, EModelElement output, ITransformationContext context)
            {
            }
            
            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EObjectTransformation>());
            }
        }

        public class ENamedElementTransformation : AbstractTransformationRule<Serialization.ENamedElement, ENamedElement>
        {
            public override void Transform(Serialization.ENamedElement input, ENamedElement output, ITransformationContext context)
            {
                output.Name = input.name;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EModelElementTransformation>());
            }
        }

        public class EObjectTransformation : AbstractTransformationRule<Serialization.EObject, EObject>
        {
            public override void Transform(Serialization.EObject input, EObject output, ITransformationContext context)
            {
                output.Id = input.id;
            }
        }

        public class EOperationTransformation : TransformationRule<Serialization.EOperation, EOperation>
        {
            public override void Transform(Serialization.EOperation input, EOperation output, ITransformationContext context)
            {
                if (input.eExceptions != null)
                {
                    var types = input.eExceptions.Split(' ');
                    foreach (var type in types)
                    {
                        var eClass = (Transformation as EcoreSerializerTransformation).TraceClass(type, context);
                        if (eClass != null) output.Exceptions.Add(eClass);
                    }

                    output.Exceptions.AddRange(context.Trace.ResolveInWhere(Rule<EClassTransformation>(),
                        c => input.Items2.Contains(c)));
                }
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ETypedElementTransformation>());

                CallForEach<Serialization.EClass, EClass>(
                    cl => cl.Items1.AllOfType<Serialization.EOperation>(),
                    (cl, operations) => cl.Operations.AddRange(operations));
            }
        }

        public class EPackageTransformation : TransformationRule<Serialization.EPackage, EPackage>
        {
            public override void Transform(Serialization.EPackage input, EPackage output, ITransformationContext context)
            {
                output.NsPrefix = input.nsPrefix;
                output.NsUri = input.nsURI;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ENamedElementTransformation>());

                CallForEach<Serialization.EPackage, EPackage>(
                    pck => pck.Items1.AllOfType<Serialization.EPackage>(),
                    (pck, packages) => pck.SubPackages.AddRange(packages));
            }
        }

        public class EParameterTransformation : TransformationRule<Serialization.EParameter, EParameter>
        {
            public override void Transform(Serialization.EParameter input, EParameter output, ITransformationContext context)
            {
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ETypedElementTransformation>());

                CallForEach<Serialization.EOperation, EOperation>(
                    op => op.Items2.AllOfType<Serialization.EParameter>(),
                    (op, param) => op.Parameters.AddRange(param));
            }
        }


        public class EReferenceTransformation : TransformationRule<Serialization.EReference, EReference>
        {
            public override void Transform(Serialization.EReference input, EReference output, ITransformationContext context)
            {
                output.IsContainment = input.containment;
                output.ResolveProxies = input.resolveProxies;

                if (input.eOpposite != null)
                {
                    output.Opposite = (Transformation as EcoreSerializerTransformation).TraceReference(input.eOpposite, context);
                }
            }

            public override EReference CreateOutput(Serialization.EReference input, ITransformationContext context)
            {
                //for trace purpose
                return new EReference() { Name = input.name };
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<EStructuralFeatureTransformation>());
            }
        }

        public class EStructuralFeatureTransformation : AbstractTransformationRule<Serialization.EStructuralFeature, EStructuralFeature>
        {
            public override void Transform(Serialization.EStructuralFeature input, EStructuralFeature output, ITransformationContext context)
            {
                output.IsChangable = input.changeable;
                output.IsVolatile = input.@volatile;
                output.IsTransient = input.transient;
                output.DefaultValue = input.defaultValueLiteral;
                output.IsUnsettable = input.unsettable;
                output.IsDerived = input.derived;
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ETypedElementTransformation>());

                CallForEach<Serialization.EClass, EClass>(
                    cl => cl.Items1.AllOfType<Serialization.EStructuralFeature>(), 
                    (cl, features) => cl.Features.AddRange(features));
            }
        }

        public class ETypedElementTransformation : AbstractTransformationRule<Serialization.ETypedElement, ETypedElement>
        {
            public override void Transform(Serialization.ETypedElement input, ETypedElement output, ITransformationContext context)
            {
                output.IsOrdered = input.ordered;
                output.IsUnique = input.unique;
                output.LowerBound = input.lowerBoundSpecified ? input.lowerBound : 0;
                output.UpperBound = input.upperBoundSpecified ? input.upperBound : 1;

                if (input.eType != null)
                {
                    output.Type = (Transformation as EcoreSerializerTransformation).TraceType(input.eType, context);
                }
            }

            public override void RegisterDependencies()
            {
                MarkInstantiatingFor(Rule<ENamedElementTransformation>());
            }
        }
    }
}
