﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace NMF.Serialization.Xmi
{
    public class XmiSerializationContext : XmlSerializationContext
    {
        public XmiSerializationContext(ISerializer serializer, object root) : base(serializer, root) { }

        private Dictionary<object, string> ids = new Dictionary<object, string>();

        public string GetId(object input)
        {
            string id;
            if (!ids.TryGetValue(input, out id))
            {
                id = "$" + ids.Count;
                ids.Add(input, id);
            }
            return id;
        }

        public void SetId(object input, string id)
        {
            if (!ids.ContainsKey(input))
            {
                ids.Add(input, id);
            }
            else
            {
                ids[input] = id;
            }
        }

        private static Regex attributeRegex = new Regex(@"(@)?(?<attribute>\w+)(\.(?<index>\d+))?", RegexOptions.Compiled);

        public override object Resolve(string id, ITypeSerializationInfo type)
        {
            var obj = base.Resolve(id, type);
            if (obj == null && Root != null)
            {
                if (id.StartsWith("//"))
                {
                    var terms = id.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                    object context = Root;
                    ITypeSerializationInfo info;
                    foreach (var term in terms)
                    {
                        if (context == null) throw new InvalidOperationException();
                        info = Serializer.GetSerializationInfo(context.GetType(), false);
                        Match m = attributeRegex.Match(term);
                        if (m.Success)
                        {
                            var elementName = m.Groups["attribute"].Value;
                            var index = GetIndex(m.Groups["index"].Value);

                            var prop = GetProperty(info, elementName);
                            if (prop != null)
                            {
                                context = prop.GetValue(context, this);
                                if (index.HasValue && prop.PropertyType.IsCollection)
                                {
                                    context = GetElementAt(index.Value, context as IEnumerable);
                                }
                            }
                            else
                            {
                                throw new InvalidOperationException("The property could not be found!");
                            }
                        }
                        else
                        {
                            throw new InvalidOperationException();
                        }
                    }
                    return context;
                }
            }
            return obj;
        }

        private object GetElementAt(int index, IEnumerable context)
        {
            if (context == null) return null;
            if (context is IList)
            {
                return ((IList)context)[index];
            }
            var en = context.GetEnumerator();
            for (int i = 0; i <= index; i++)
            {
                if (!en.MoveNext()) return null;
            }
            return en.Current;
        }

        private IPropertySerializationInfo GetProperty(ITypeSerializationInfo info, string elementName)
        {
            foreach (var prop in info.ElementProperties)
            {
                if (prop.ElementName == elementName) return prop;
            }
            foreach (var prop in info.AttributeProperties)
            {
                if (prop.ElementName == elementName) return prop;
            }
            return null;
        }

        private int? GetIndex(string indexString)
        {
            if (string.IsNullOrEmpty(indexString))
            {
                return null;
            }
            else
            {
                return int.Parse(indexString);
            }
        }
    }
}
