﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace NMF.Serialization.Xmi
{
    public class XmiSerializer : XmlSerializer
    {
        /// <summary>
        /// Creates a new XmiSerializer with default settings and no preloaded types
        /// </summary>
        public XmiSerializer() : this(XmlSerializationSettings.Default) { }

        /// <summary>
        /// Creates a new XmiSerializer with default settings
        /// </summary>
        /// <param name="additionalTypes">Set of types to preload into the serializer</param>
        /// <remarks>Types will be loaded with default settings</remarks>
        public XmiSerializer(IEnumerable<Type> additionalTypes) : this(XmlSerializationSettings.Default, additionalTypes) { }

        /// <summary>
        /// Creates a new XmiSerializer with the specified settings
        /// </summary>
        /// <param name="settings">Serializer-settings for the serializer. Can be null or Nothing in Visual Basic. In this case, the default settings will be used.</param>
        public XmiSerializer(XmlSerializationSettings settings) : base(settings) {}

        /// <summary>
        /// Creates a new XmiSerializer with the specified settings and the given preloaded types
        /// </summary>
        /// <param name="additionalTypes">Set of types to load into the serializer</param>
        /// <param name="settings">The settings to use for the serializer</param>
        /// <remarks>The types will be loaded with the specified settings</remarks>
        public XmiSerializer(XmlSerializationSettings settings, IEnumerable<Type> additionalTypes) : base(settings, additionalTypes) { }

        private static readonly string TypeField = "type";
        public static readonly string XMLSchemaInstanceNamespace = "http://www.w3.org/2001/XMLSchema-instance";
        public static readonly string XMLSchemaInstancePrefix = "xsi";

        public static readonly string XMIPrefix = "xmi";
        public static readonly string XMINamespace = "http://www.omg.org/XMI";

        protected override bool GoToPropertyContent(System.Xml.XmlReader reader)
        {
            //do not move forward
            return true;
        }

        protected override ISerializationContext CreateSerializationContext(object root)
        {
            return new XmiSerializationContext(this, root);
        }

        protected override bool IsElementOfType(System.Xml.XmlReader reader, IPropertySerializationInfo property, ITypeSerializationInfo tsi)
        {
            string attr = reader.GetAttribute(TypeField, XMLSchemaInstanceNamespace);
            if (attr != null)
            {
                int separator = attr.IndexOf(':');
                if (separator == -1) return attr == tsi.ElementName;
                return tsi.NamespacePrefix == attr.Substring(0, separator) && tsi.ElementName == attr.Substring(separator + 1);
            }
            else
            {
                if (property == null) throw new InvalidOperationException();
                if (property.PropertyType.IsCollection)
                {
                    return tsi == property.PropertyType.CollectionItemType;
                }
                else
                {
                    return tsi == property.PropertyType;
                }
            }
        }

        protected override void WriteBeginElement(System.Xml.XmlWriter writer, object obj, ITypeSerializationInfo info)
        {
            string prefix = info.NamespacePrefix;
            if (prefix == null) prefix = writer.LookupPrefix(info.Namespace);

            string value = prefix + ":" + info.ElementName;
            writer.WriteAttributeString(XMLSchemaInstancePrefix, TypeField, XMLSchemaInstanceNamespace, value);
        }

        protected override void WriteEndElement(System.Xml.XmlWriter writer, object obj, ITypeSerializationInfo info)
        {
        }

        protected override void WriteBeginRootElement(System.Xml.XmlWriter writer, object root, ITypeSerializationInfo info)
        {
            base.WriteBeginElement(writer, root, info);

            writer.WriteAttributeString(XMIPrefix, "version", XMINamespace, "2.0"); 
        }

        protected override void WriteEndRootElement(System.Xml.XmlWriter writer, object root, ITypeSerializationInfo info)
        {
            writer.WriteEndElement();
        }

        protected override void WriteElementProperties(System.Xml.XmlWriter writer, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            foreach (IPropertySerializationInfo pi in info.ElementProperties)
            {
                if (pi.ShouldSerializeValue(obj))
                {
                    if (pi.PropertyType.IsCollection)
                    {
                        var collectionType = pi.PropertyType.CollectionItemType.Type;
                        foreach (object item in pi.GetValue(obj, context) as IEnumerable)
                        {
                            writer.WriteStartElement(pi.NamespacePrefix, pi.ElementName, pi.Namespace);
                            var itemType = item.GetType();
                            if (itemType != collectionType)
                            {
                                var itemInfo = GetSerializationInfo(itemType, false);
                                writer.WriteAttributeString(XMLSchemaInstancePrefix, TypeField, XMLSchemaInstanceNamespace,
                                    itemInfo.NamespacePrefix + ":" + itemInfo.ElementName);
                            }
                            Serialize(item, writer, pi, false, pi.IdentificationMode, context);
                            writer.WriteEndElement();
                        }
                    }
                    else
                    {
                        var target = pi.GetValue(obj, context);
                        if (target != null)
                        {
                            var targetType = target.GetType();
                            writer.WriteStartElement(pi.NamespacePrefix, pi.ElementName, pi.Namespace);
                            if (pi.PropertyType.Type != targetType)
                            {
                                var targetInfo = GetSerializationInfo(targetType, false);
                                writer.WriteAttributeString(XMLSchemaInstancePrefix, TypeField, XMLSchemaInstanceNamespace,
                                    targetInfo.NamespacePrefix + ":" + targetInfo.ElementName);
                            }
                            Serialize(pi.GetValue(obj, context), writer, pi, false, pi.IdentificationMode, context);
                            writer.WriteEndElement();
                        }
                    }
                }
            }
        }

        protected override void InitializeElementProperties(System.Xml.XmlReader reader, ref object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            int currentDepth = reader.Depth;
            while (reader.Depth < currentDepth || reader.Read())
            {
                if (reader.Depth == currentDepth)
                {
                    break;
                }
                else if (reader.Depth < currentDepth)
                {
                    return;
                }
                if (reader.NodeType == XmlNodeType.Element)
                {
                    foreach (IPropertySerializationInfo p in info.ElementProperties)
                    {
                        if (IsPropertyElement(reader, p))
                        {
                            object current = DeserializeInternal(reader, p, context);
                            if (p.PropertyType.IsCollection)
                            {
                                p.PropertyType.AddToCollection(p.GetValue(obj, context), current);
                            }
                            else
                            {
                                p.SetValue(obj, current, context);
                            }
                            break;
                        }
                    }
                }
            }
        }

        protected override void InitializeTypeSerializationInfo(Type type, ITypeSerializationInfo serializationInfo)
        {
            base.InitializeTypeSerializationInfo(type, serializationInfo);
            if (serializationInfo is XmlTypeSerializationInfo)
            {
                XmlTypeSerializationInfo info = serializationInfo as XmlTypeSerializationInfo;
                if (!serializationInfo.IsIdentified && !serializationInfo.IsCollection)
                {
                    info.AttributeProperties.Add(XmiArtificialIdAttribute.Instance);
                    info.IdentifierProperty = XmiArtificialIdAttribute.Instance;
                }
                info.CheckElementsExist = true;
            }
        }
    }
}
