﻿using System.Reflection;
using System.ComponentModel;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Collections.ObjectModel;
using System.Xml;
using System.Text;

namespace NMF.Serialization
{

    /// <summary>
    /// Class to serialize objects in a Xml-format.
    /// </summary>
    public class XmlSerializer : ISerializer
    {
        private Dictionary<Type, ITypeSerializationInfo> types = new Dictionary<Type, ITypeSerializationInfo>();
        private XmlTypeCollection typesWrapper;

        private static Type genericCollection = typeof(ICollection<>);

        private XmlSerializationSettings settings;
        private Queue<XmlIdentifierDelay> lostProperties = new Queue<XmlIdentifierDelay>();
        private Queue<ISupportInitialize> inits = new Queue<ISupportInitialize>();

        private object[] emptyObjects = {};

        /// <summary>
        /// Creates a new XmlSerializer with default settings and no preloaded types
        /// </summary>
        public XmlSerializer() : this(XmlSerializationSettings.Default) { }

        /// <summary>
        /// Creates a new XmlSerializer with default settings
        /// </summary>
        /// <param name="additionalTypes">Set of types to preload into the serializer</param>
        /// <remarks>Types will be loaded with default settings</remarks>
        public XmlSerializer(IEnumerable<Type> additionalTypes) : this(XmlSerializationSettings.Default, additionalTypes) { }

        /// <summary>
        /// Creates a new XmlSerializer with the specified settings
        /// </summary>
        /// <param name="settings">Serializer-settings for the serializer. Can be null or Nothing in Visual Basic. In this case, the default settings will be used.</param>
        public XmlSerializer(XmlSerializationSettings settings)
        {
            this.settings = settings;
            if (settings == null) settings = XmlSerializationSettings.Default;
            this.typesWrapper = new XmlTypeCollection(this);
        }

        /// <summary>
        /// Creates a new XmlSerializer with the specified settings and the given preloaded types
        /// </summary>
        /// <param name="additionalTypes">Set of types to load into the serializer</param>
        /// <param name="settings">The settings to use for the serializer</param>
        /// <remarks>The types will be loaded with the specified settings</remarks>
        public XmlSerializer(XmlSerializationSettings settings, IEnumerable<Type> additionalTypes)
        {
            this.settings = settings;
            if (settings == null) settings = XmlSerializationSettings.Default;
            this.typesWrapper = new XmlTypeCollection(this);
            foreach (Type t in additionalTypes)
            {
                AddType(t);
            }
        }

        /// <summary>
        /// The settings to be used in the serializer
        /// </summary>
        public XmlSerializationSettings Settings
        {
            get { return settings; }
        }

        /// <summary>
        /// The set of types that are known to the serializer
        /// </summary>
        public ICollection<Type> KnownTypes
        {
            get
            {
                return typesWrapper;
            }
        }

        internal Dictionary<Type, ITypeSerializationInfo> Types
        {
            get { return types; }
        }

        internal void AddType(Type type)
        {
            ITypeSerializationInfo info = CreateTypeSerializationInfoFor(type);
            types.Add(type, info);

            InitializeTypeSerializationInfo(type, info);
        }

        protected virtual ITypeSerializationInfo CreateTypeSerializationInfoFor(Type type)
        {
            return new XmlTypeSerializationInfo(type);
        }

        protected virtual void InitializeTypeSerializationInfo(Type type, ITypeSerializationInfo serializationInfo)
        {
            XmlTypeSerializationInfo info = serializationInfo as XmlTypeSerializationInfo;

            if (info == null) throw new NotSupportedException("Cannot initialize other serialization info types");

            string identifier = null;
            XmlPropertySerializationInfo[] constructorInfos = null;
            info.ElementName = Settings.GetPersistanceString(type.Name);
            info.Namespace = Settings.DefaultNamespace;

            foreach (object att in type.GetCustomAttributes(typeof(XmlElementNameAttribute), false))
            {
                info.ElementName = Settings.GetPersistanceString((att as XmlElementNameAttribute).ElementName);
            }

            foreach (object att in type.GetCustomAttributes(typeof(XmlNamespaceAttribute), false))
            {
                info.Namespace = (att as XmlNamespaceAttribute).Namespace;
            }

            foreach (object att in type.GetCustomAttributes(typeof(XmlNamespacePrefixAttribute), false))
            {
                info.NamespacePrefix = (att as XmlNamespacePrefixAttribute).NamespacePrefix;
            }

            foreach (object att in type.GetCustomAttributes(typeof(XmlConstructorAttribute), false))
            {
                constructorInfos = new XmlPropertySerializationInfo[(att as XmlConstructorAttribute).ParameterCount];
                info.ConstructorProperties = constructorInfos;
            }

            foreach (object att in type.GetCustomAttributes(typeof(XmlIdentifierAttribute), false))
            {
                identifier = (att as XmlIdentifierAttribute).Identifier;
            }

            List<string> ignoredProperties = new List<string>();
            foreach (object att in type.GetCustomAttributes(typeof(XmlIgnorePropertyAttribute), false))
            {
                ignoredProperties.Add((att as XmlIgnorePropertyAttribute).Property);
            }

            if (typeof(IEnumerable).IsAssignableFrom(type))
            {
                foreach (Type i in type.GetInterfaces())
                {
                    if (i.IsGenericType && i.GetGenericTypeDefinition() == genericCollection)
                    {
                        Type collType = i.GetGenericArguments()[0];
                        info.CollectionType = i;
                        info.CollectionItemType = GetSerializationInfo(collType, true);
                        break;
                    }
                }
            }
            foreach (PropertyDescriptor pd in TypeDescriptor.GetProperties(type))
            {
                RegisterPropertySerializationInfo(info, identifier, constructorInfos, ignoredProperties, pd);
            }
            Type constructorType = type;
            if (type.IsInterface)
            {
                foreach (var implementedInterface in type.GetInterfaces())
                {
                    ITypeSerializationInfo parentTsi = GetSerializationInfo(implementedInterface, true);
                    info.BaseTypes.Add(parentTsi);
                    foreach (var attProp in parentTsi.AttributeProperties)
                    {
                        if (!info.AttributeProperties.Contains(attProp))
                        {
                            info.AttributeProperties.Add(attProp);
                            if (attProp.IsIdentifier) info.IdentifierProperty = attProp;
                        }
                    }
                    foreach (var elProp in parentTsi.ElementProperties)
                    {
                        if (!info.ElementProperties.Contains(elProp))
                        {
                            info.ElementProperties.Add(elProp);
                            if (elProp.IsIdentifier) info.IdentifierProperty = elProp;
                        }
                    }
                }
                foreach (object att in type.GetCustomAttributes(typeof(XmlDefaultImplementationTypeAttribute), false))
                {
                    constructorType = (att as XmlDefaultImplementationTypeAttribute).DefaultImplementationType;
                    if (constructorType == null)
                    {
                        constructorType = type;
                    }
                    else
                    {
                        types.Add(constructorType, info);
                    }
                }
            }
            else
            {
                foreach (var implementedInterface in type.GetInterfaces())
                {
                    ITypeSerializationInfo parentTsi = GetSerializationInfo(implementedInterface, true);
                    info.BaseTypes.Add(parentTsi);
                }
                if (type.BaseType != null)
                {
                    var parentTsi = GetSerializationInfo(type.BaseType, true);
                    info.BaseTypes.Add(parentTsi);
                }
            }
            if (constructorInfos != null)
            {
                Type[] ts = new Type[constructorInfos.GetLength(0)];
                for (int i = 0; i < constructorInfos.GetLength(0); i++)
                {
                    ts[i] = constructorInfos[i] == null ? typeof(object) : constructorInfos[i].Descriptor.PropertyType;
                }
                info.Constructor = constructorType.GetConstructor(ts);
                if (info.Constructor == null) throw new InvalidOperationException("No suitable constructor found for type " + type.FullName);
            }
            else
            {
                info.Constructor = constructorType.GetConstructor(Type.EmptyTypes);
            }
            foreach (object att in type.GetCustomAttributes(typeof(XmlKnownTypeAttribute), false))
            {
                var t = (att as XmlKnownTypeAttribute).Type;
                if (t != null && !types.ContainsKey(t)) AddType(t);
            }
        }

        protected virtual void RegisterPropertySerializationInfo(XmlTypeSerializationInfo typeSerializationInfo, string identifier, XmlPropertySerializationInfo[] constructorInfos, List<string> ignoredProperties, PropertyDescriptor pd)
        {
            if (ignoredProperties.Contains(pd.Name)) return;

            XmlPropertySerializationInfo p;
            p = CreatePropertySerializationInfo(pd);

            DesignerSerializationVisibilityAttribute des = pd.Attributes[typeof(DesignerSerializationVisibilityAttribute)] as DesignerSerializationVisibilityAttribute;
            object att;
            att = pd.Attributes[typeof(XmlConstructorParameterAttribute)];

            if ((des == null || des.Visibility == DesignerSerializationVisibility.Visible) && !p.Descriptor.IsReadOnly)
            {
                p.ShallCreateInstance = true;
            }
            else if (des != null && des.Visibility == DesignerSerializationVisibility.Content)
            {
                p.ShallCreateInstance = false;

            }
            else
            {
                if (att == null || des.Visibility == DesignerSerializationVisibility.Hidden)
                {
                    return;
                }
                else
                {
                    p.ShallCreateInstance = false;
                }
            }

            if (pd.Name == identifier)
            {
                p.IsIdentifier = true;
                typeSerializationInfo.IdentifierProperty = p;
            }

            //auch hier soll gesteuert werden, ob die Eigenschaft als Attribut gerendert werden soll
            if (att != null && constructorInfos != null)
            {
                XmlConstructorParameterAttribute pa = att as XmlConstructorParameterAttribute;
                if (pa.Index >= 0 || pa.Index < constructorInfos.GetLength(0))
                {
                    constructorInfos[pa.Index] = p;
                }
                else
                {
                    att = pd.Attributes[typeof(XmlAttributeAttribute)];
                    if (att == null || !(att as XmlAttributeAttribute).SerializeAsAttribute)
                    {
                        typeSerializationInfo.ElementProperties.Add(p);
                    }
                    else
                    {
                        typeSerializationInfo.AttributeProperties.Add(p);
                    }
                }
            }
            else
            {
                att = pd.Attributes[typeof(XmlAttributeAttribute)];
                if (att == null || !(att as XmlAttributeAttribute).SerializeAsAttribute)
                {
                    typeSerializationInfo.ElementProperties.Add(p);
                }
                else
                {
                    typeSerializationInfo.AttributeProperties.Add(p);
                }
            }


            //Defaulteinstellungen, können durch Attribute überschrieben werden
            p.ElementName = Settings.GetPersistanceString(pd.Name);
            p.Namespace = Settings.DefaultNamespace;
            //
            att = pd.Attributes[typeof(XmlElementNameAttribute)];
            if (att != null) p.ElementName = Settings.GetPersistanceString((att as XmlElementNameAttribute).ElementName);
            att = pd.Attributes[typeof(XmlNamespaceAttribute)];
            if (att != null) p.Namespace = Settings.GetPersistanceString((att as XmlNamespaceAttribute).Namespace);
            att = pd.Attributes[typeof(XmlIdentificationModeAttribute)];
            if (att != null) p.IdentificationMode = (att as XmlIdentificationModeAttribute).Mode;
            att = pd.Attributes[typeof(XmlNamespacePrefixAttribute)];
            if (att != null) p.NamespacePrefix = Settings.GetPersistanceString((att as XmlNamespacePrefixAttribute).NamespacePrefix);


            //evtl benutzt die property einen eigenen TypeConverter
            att = pd.Attributes[typeof(XmlTypeConverterAttribute)];
            if (att != null)
            {
                try
                {
                    p.CustomConverter = Activator.CreateInstance((att as XmlTypeConverterAttribute).Type) as TypeConverter;
                }
                catch (Exception)
                { }
            }

            if (p.Converter == null || !p.Converter.CanConvertFrom(typeof(string)) || !p.Converter.CanConvertTo(typeof(string)))
            {
                if (!types.ContainsKey(pd.PropertyType)) AddType(pd.PropertyType);
                p.PropertyType = types[pd.PropertyType];
            }
        }

        protected virtual XmlPropertySerializationInfo CreatePropertySerializationInfo(PropertyDescriptor pd)
        {
            return new XmlPropertySerializationInfo(pd);
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="path">The path for the resulting Xml-file</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object obj, string path)
        {
            using (FileStream fs = new FileStream(path, FileMode.Create, FileAccess.ReadWrite, FileShare.None))
            {
                Serialize(obj, fs);
            }
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="stream">The stream for the resulting Xml-code</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object source, Stream stream)
        {
            using (var sw = new StreamWriter(stream, Encoding.UTF8))
            {
                Serialize(source, sw);
            }
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="writer">The TextWriter to write the Xml-code on</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object source, TextWriter writer)
        {
            Serialize(source, writer, false);
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="writer">The XmlWriter to write the Xml-code on</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object source, XmlWriter writer)
        {
            Serialize(source, writer, false);
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="target">The TextWriter to write the Xml-code on</param>
        /// <param name="fragment">A value that indicates whether the serializer should write a document definition</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object source, TextWriter target, bool fragment)
        {
            XmlWriter xml = XmlWriter.Create(target, Settings.GetXmlWriterSettings());
            Serialize(source, xml, fragment);
            xml.Flush();
            xml.Close();
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="target">The XmlWriter to write the Xml-code on</param>
        /// <param name="fragment">A value that indicates whether the serializer should write a document definition</param>
        /// <param name="o">The object to be serialized</param>
        public void Serialize(object source, XmlWriter target, bool fragment)
        {
            if (!fragment) target.WriteStartDocument();
            if (!types.ContainsKey(source.GetType())) AddType(source.GetType());
            ITypeSerializationInfo info = types[source.GetType()];
            WriteBeginRootElement(target, source, info);
            ISerializationContext context = CreateSerializationContext(source);
            Serialize(source, target, null, false, XmlIdentificationMode.FullObject, context);
            WriteEndRootElement(target, source, info);
            if (!fragment) target.WriteEndDocument();
        }

        protected virtual ISerializationContext CreateSerializationContext(object root)
        {
            return new XmlSerializationContext(this, root);
        }

        /// <summary>
        /// Serializes the given object
        /// </summary>
        /// <param name="writer">The XmlWriter to write the Xml-code on</param>
        /// <param name="converter">A TypeConverter that might convert the object straight to string. Can be left out.</param>
        /// <param name="writeInstance">A value that indicates whether the serializer should write the element definition</param>
        /// <param name="o">The object to be serialized</param>
        /// <param name="identificationMode">A value indicating whether it is allowed to the serializer to use identifier</param>
        /// <remarks>If a converter is provided that is able to convert the object to string and convert the string back to this object, just the string-conversion is printed out</remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public void Serialize(object obj, XmlWriter writer, IPropertySerializationInfo property, bool writeInstance, XmlIdentificationMode identificationMode, ISerializationContext context)
        {
            if (obj == null) return;
            if (property != null && property.IsStringConvertible)
            {
                writer.WriteString(property.ConvertToString(obj));
                return;
            }
            ITypeSerializationInfo info = GetSerializationInfo(obj.GetType(), false);
            if (info.IsIdentified)
            {
                if (WriteIdentifiedObject(writer, obj, identificationMode, info, context)) return;
            }
            if (writeInstance) WriteBeginElement(writer, obj, info);
            if (info.ConstructorProperties != null)
            {
                WriteConstructorProperties(writer, obj, info, context);
            }
            WriteAttributeProperties(writer, obj, info, context);
            WriteElementProperties(writer, obj, info, context);
            WriteCollectionMembers(writer, obj, info, context);
            if (writeInstance) WriteEndElement(writer, obj, info);
        }

        protected virtual void WriteBeginRootElement(XmlWriter writer, object root, ITypeSerializationInfo info)
        {
            WriteBeginElement(writer, root, info);
        }

        protected virtual void WriteBeginElement(XmlWriter writer, object obj, ITypeSerializationInfo info)
        {
            writer.WriteStartElement(info.NamespacePrefix, info.ElementName, info.Namespace);
        }

        protected virtual void WriteConstructorProperties(XmlWriter writer, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            for (int i = 0; i <= info.ConstructorProperties.GetUpperBound(0); i++)
            {
                IPropertySerializationInfo pi = info.ConstructorProperties[i];
                writer.WriteAttributeString(pi.NamespacePrefix, pi.ElementName, pi.Namespace, pi.ConvertToString(pi.GetValue(obj, context)));
            }
        }

        protected virtual void WriteAttributeProperties(XmlWriter writer, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            foreach (IPropertySerializationInfo pi in info.AttributeProperties)
            {
                if (pi.ShouldSerializeValue(obj)) WriteAttributeValue(writer, obj, pi, context);
                if (pi.IsIdentifier)
                {
                    string id = CStr(pi.GetValue(obj, context));
                    context.RegisterId(id, obj);
                }
            }
        }

        protected virtual void WriteAttributeValue(XmlWriter writer, object obj, IPropertySerializationInfo property, ISerializationContext context)
        {
            ITypeSerializationInfo info = property.PropertyType;
            object reference = property.GetValue(obj, context);

            if (reference == null) return;

            if (property.IsStringConvertible)
            {
                writer.WriteAttributeString(property.NamespacePrefix, property.ElementName, property.Namespace, property.ConvertToString(reference));
            }
            else if (info.IsIdentified)
            {
                string id = CStr(info.IdentifierProperty.GetValue(reference, context));
                writer.WriteAttributeString(property.NamespacePrefix, property.ElementName, property.Namespace, id);
            }
            else if (info.IsCollection)
            {
                info = info.CollectionItemType;
                StringBuilder sb = new StringBuilder();
                foreach (object o in reference as IEnumerable)
                {
                    if (info.IsIdentified)
                    {
                        string id = CStr(info.IdentifierProperty.GetValue(o, context));
                        // do not register id! Otherwise, the next time the item is serialized, it is serialized as id, only
                        sb.Append(id);
                        sb.Append(" ");
                    }
                    else if (info.IsStringConvertible)
                    {
                        sb.Append(info.ConvertToString(o));
                        sb.Append(" ");
                    }
                    else
                    {
                        throw new InvalidOperationException(string.Format("Property {0} cannot be serialized as string", property.ElementName));
                    }
                }
                if (sb.Length > 0) sb.Remove(sb.Length - 1, 1);
                writer.WriteAttributeString(property.NamespacePrefix, property.ElementName, property.Namespace, sb.ToString());
            }
            else
            {
                throw new InvalidOperationException(string.Format("Property {0} cannot be serialized as string", property.ElementName));
            }
        }

        protected virtual void WriteElementProperties(XmlWriter writer, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            foreach (IPropertySerializationInfo pi in info.ElementProperties)
            {
                if (pi.ShouldSerializeValue(obj))
                {
                    writer.WriteStartElement(pi.NamespacePrefix, pi.ElementName, pi.Namespace);
                    Serialize(pi.GetValue(obj, context), writer, pi, pi.ShallCreateInstance, pi.IdentificationMode, context);
                    writer.WriteEndElement();
                }
                if (pi.IsIdentifier)
                {
                    string id = CStr(pi.GetValue(obj, context));
                    context.RegisterId(id, obj);
                }
            }
        }

        protected virtual void WriteCollectionMembers(XmlWriter writer, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            if (info.IsCollection)
            {
                IEnumerable coll = obj as IEnumerable;
                foreach (object o in coll)
                {
                    Serialize(o, writer, null, true, XmlIdentificationMode.FullObject, context);
                }
            }
        }

        protected virtual void WriteEndElement(XmlWriter writer, object obj, ITypeSerializationInfo info)
        {
            writer.WriteEndElement();
        }

        protected virtual void WriteEndRootElement(XmlWriter writer, object root, ITypeSerializationInfo info)
        {
            writer.WriteEndElement();
        }

        protected virtual bool WriteIdentifiedObject(XmlWriter writer, object obj, XmlIdentificationMode identificationMode, ITypeSerializationInfo info, ISerializationContext context)
        {
            string id = CStr(info.IdentifierProperty.GetValue(obj, context));
            if (identificationMode == XmlIdentificationMode.Identifier || (identificationMode == XmlIdentificationMode.AsNeeded && context.ContainsId(id, info)))
            {
                writer.WriteString(id);
                return true;
            }
            else if (identificationMode == XmlIdentificationMode.FullObject && context.ContainsId(id, info))
            {
                writer.WriteStartElement(info.ElementName, info.Namespace);
                if (info.AttributeProperties.Contains(info.IdentifierProperty))
                {
                    writer.WriteAttributeString(info.IdentifierProperty.NamespacePrefix, info.IdentifierProperty.ElementName, info.IdentifierProperty.Namespace,
                        info.IdentifierProperty.ConvertToString(info.IdentifierProperty.GetValue(obj, context)));
                }
                else
                {
                    writer.WriteElementString(info.IdentifierProperty.NamespacePrefix, info.IdentifierProperty.ElementName, info.IdentifierProperty.Namespace,
                        info.IdentifierProperty.ConvertToString(info.IdentifierProperty.GetValue(obj, context)));
                }
                writer.WriteEndElement();
                return true;
            }
            return false;
        }

        private static string CStr(object obj)
        {
            return obj == null ? null : obj.ToString();
        }
         
        /// <summary>
        /// Deserializes an Xml-representation of an object back to the corresponding object
        /// </summary>
        /// <param name="path">The path to the Xml file containg the Xml code</param>
        /// <returns>The corresponding object</returns>
        public object Deserialize(string path)
        {
            using (FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                return Deserialize(fs);
            }
        }

        /// <summary>
        /// Deserializes an Xml-representation of an object back to the corresponding object
        /// </summary>
        /// <param name="stream">The stream containg the Xml code</param>
        /// <returns>The corresponding object</returns>
        public object Deserialize(Stream stream)
        {
            return Deserialize(XmlReader.Create(stream));
        }

        /// <summary>
        /// Deserializes an Xml-representation of an object back to the corresponding object
        /// </summary>
        /// <param name="reader">A TextReader containg the Xml code</param>
        /// <returns>The corresponding object</returns>
        public object Deserialize(TextReader reader)
        {
            return Deserialize(XmlReader.Create(reader));
        }

        /// <summary>
        /// Deserializes an Xml-representation of an object back to the corresponding object
        /// </summary>
        /// <param name="reader">A XmlReader containing the Xml code</param>
        /// <returns>The corresponding object</returns>
        /// <remarks>The function will deserialize the object at the XmlReaders current position</remarks>
        public object Deserialize(XmlReader reader)
        {
            ISerializationContext context;
            inits.Clear();
            lostProperties.Clear();
            object root = DeserializeRootInternal(reader, out context);
            while (lostProperties.Count > 0)
            {
                XmlIdentifierDelay p = lostProperties.Dequeue();
                p.OnResolveIdentifiedObject(context.Resolve(p.Identifier, p.TargetType), context);
            }
            while (inits.Count > 0)
            {
                inits.Dequeue().EndInit();
            }
            return root;
        }
        
        protected object DeserializeInternal(XmlReader reader, IPropertySerializationInfo property, ISerializationContext context)
        {
            object d = null;
            while (reader.NodeType != XmlNodeType.Element) reader.Read();
            if (property != null && IsElementOfType(reader, property, property.PropertyType))
            {
                d = CreateObject(reader, property.PropertyType);
            }
            else
            {
                foreach (ITypeSerializationInfo tsi in types.Values)
                {
                    if (IsElementOfType(reader, property, tsi))
                    {
                        d = CreateObject(reader, tsi);
                        break;
                    }
                }
            }
            if (d == null)
            {
                throw new InvalidDataException(
                    string.Format("The type {0} in the namespace {1} could not be matched. ",
                    reader.Name, reader.NamespaceURI));
            }
            Initialize(reader, d, context);
            return d;
        }

        protected object DeserializeRootInternal(XmlReader reader, out ISerializationContext context)
        {
            object root = null;
            while (reader.NodeType != XmlNodeType.Element) reader.Read();
            foreach (ITypeSerializationInfo tsi in types.Values)
            {
                if (IsRootElementOfType(reader, tsi))
                {
                    root = CreateObject(reader, tsi);
                    break;
                }
            }
            if (root == null) throw new ArgumentNullException(
                string.Format("The type {0} in the namespace {1} could not be matched. ",
                reader.Name, reader.NamespaceURI));

            context = CreateSerializationContext(root);

            Initialize(reader, root, context);
            return root;
        }

        protected virtual bool IsElementOfType(XmlReader reader, IPropertySerializationInfo property, ITypeSerializationInfo type)
        {
            return Settings.EqualizeStrings(type.ElementName, reader.LocalName) && Settings.EqualizeStrings(type.Namespace, reader.NamespaceURI);
        }

        protected virtual bool IsRootElementOfType(XmlReader reader, ITypeSerializationInfo tsi)
        {
            return Settings.EqualizeStrings(tsi.ElementName, reader.LocalName) && Settings.EqualizeStrings(tsi.Namespace, reader.NamespaceURI);
        }

        protected virtual object CreateObject(XmlReader reader, ITypeSerializationInfo tsi)
        {
            if (tsi.ConstructorProperties == null)
            {
                return tsi.Constructor.Invoke(emptyObjects);
            }
            else
            {
                object[] objects = new object[tsi.ConstructorProperties.Length];
                for (int i = 0; i < tsi.ConstructorProperties.Length; i++)
                {
                    IPropertySerializationInfo pi = tsi.ConstructorProperties[i];
                    objects[i] = pi.ConvertFromString(reader.GetAttribute(pi.ElementName, pi.Namespace));
                }
                return tsi.Constructor.Invoke(objects);
            }
        }

        protected virtual bool InitializeProperty(XmlReader reader, IPropertySerializationInfo property, object obj, ISerializationContext context)
        {
            if (!GoToPropertyContent(reader)) return false;
            if (reader.NodeType == XmlNodeType.Text)
            {
                InitializePropertyFromText(property, obj, reader.Value, context);
            }
            else if (reader.NodeType != XmlNodeType.EndElement)
            {
                object target = DeserializeInternal(reader, property, context);
                if (!property.IsReadOnly && (target == null || property.PropertyType.Type.IsAssignableFrom(target.GetType())))
                {
                    property.SetValue(obj, target, context);
                }
                else if (property.PropertyType.IsCollection)
                {
                    object collection = property.GetValue(obj, context);
                    property.PropertyType.AddToCollection(collection, target);
                }
            }
            else
            {
                //do nothing
            }
            return true;
        }

        protected virtual bool GoToPropertyContent(XmlReader reader)
        {
            int currentDepth = reader.Depth;
            reader.Read();
            if (reader.Depth <= currentDepth) return false;
            return true;
        }

        protected virtual void InitializePropertyFromText(IPropertySerializationInfo property, object obj, string text, ISerializationContext context)
        {
            ITypeSerializationInfo info = property.PropertyType;
            if (property.IsStringConvertible)
            {
                property.SetValue(obj, property.ConvertFromString(text), context);
            }
            else if (info.IsCollection)
            {
                ITypeSerializationInfo itemInfo = info.CollectionItemType;
                object coll = property.GetValue(obj, context);
                foreach (var item in text.Split(new char[] { ' '}, StringSplitOptions.RemoveEmptyEntries))
                {
                    if (itemInfo.IsStringConvertible)
                    {
                        info.AddToCollection(coll, itemInfo.ConvertFromString(item));
                    }
                    else
                    {
                        EnqueueAddToCollectionDelay(info, item, coll);
                    }
                }
            }
            else
            {
                EnqueueSetPropertyDelay(property, obj, text);
            }
        }

        private void EnqueueSetPropertyDelay(IPropertySerializationInfo property, object obj, string text)
        {
            lostProperties.Enqueue(new XmlSetPropertyDelay() { Identifier = text, Target = obj, Property = property });
        }

        private void EnqueueAddToCollectionDelay(ITypeSerializationInfo info, string item, object coll)
        {
            lostProperties.Enqueue(new XmlAddToCollectionDelay() { Identifier = item, Target = coll, Type = info });
        }

        /// <summary>
        /// Initializes the given object with the xml code at the current position of the XmlReader
        /// </summary>
        /// <param name="reader">The XmlReader with the Xml code</param>
        /// <param name="o">The object to initialize</param>
        /// <returns>The initialized object</returns>
        public void Initialize(XmlReader reader, object obj, ISerializationContext context)
        {
            if (obj == null) return;
            if (obj is ISupportInitialize) ((ISupportInitialize)obj).BeginInit();
            ITypeSerializationInfo info = GetSerializationInfo(obj.GetType(), false);
            if (reader.HasAttributes)
            {
                if (info.IsIdentified && info.AttributeProperties.Contains(info.IdentifierProperty))
                {
                    IPropertySerializationInfo p = info.IdentifierProperty;
                    string id = CStr(p.ConvertFromString(reader.GetAttribute(p.ElementName, p.Namespace)));
                    if (!string.IsNullOrEmpty(id))
                    {
                        if (!context.ContainsId(id, info))
                        {
                            context.RegisterId(id, obj);
                        }
                        else
                        {
                            obj = context.Resolve(id, info);
                        }
                    }
                }
                InitializeAttributeProperties(reader, obj, info, context);
            }
            InitializeElementProperties(reader, ref obj, info, context);
            if (obj is ISupportInitialize) inits.Enqueue(obj as ISupportInitialize);
        }

        protected virtual void InitializeElementProperties(XmlReader reader, ref object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            int currentDepth = reader.Depth;
            bool found;
            while (reader.Depth < currentDepth || reader.Read())
            {
                if (reader.Depth == currentDepth)
                {
                    break;
                }
                else if (reader.Depth < currentDepth)
                {
                    return;
                }
                if (reader.NodeType == XmlNodeType.Element)
                {
                    found = false;
                    foreach (IPropertySerializationInfo p in info.ElementProperties)
                    {
                        if (IsPropertyElement(reader, p))
                        {
                            if (p.ShallCreateInstance)
                            {
                                if (!InitializeProperty(reader, p, obj, context))
                                {
                                    found = true;
                                    break;
                                }
                            }
                            else
                            {
                                Initialize(reader, p.GetValue(obj, context), context);
                            }
                            if (p.IsIdentifier)
                            {
                                string str = CStr(p.GetValue(obj, context));
                                if (!string.IsNullOrEmpty(str))
                                {
                                    if (context.ContainsId(str, info))
                                    {
                                        obj = context.Resolve(str, info);
                                    }
                                    else
                                    {
                                        context.RegisterId(str, obj);
                                    }
                                }
                            }
                            found = true;
                            break;
                        }
                    }
                    if (!found && info.IsCollection)
                    {
                        object o = DeserializeInternal(reader, null, context);
                        info.AddToCollection(obj, o);
                    }

                }
            }
        }

        protected virtual bool IsPropertyElement(XmlReader reader, IPropertySerializationInfo p)
        {
            return Settings.EqualizeStrings(reader.NamespaceURI, p.Namespace) && Settings.EqualizeStrings(reader.LocalName, p.ElementName);
        }

        protected virtual void InitializeAttributeProperties(XmlReader reader, object obj, ITypeSerializationInfo info, ISerializationContext context)
        {
            foreach (IPropertySerializationInfo p in info.AttributeProperties)
            {
                string attr = reader.GetAttribute(p.ElementName, p.Namespace);
                if (attr != null)
                {
                    InitializePropertyFromText(p, obj, attr, context);
                }
            }
        }

        protected virtual bool HandleException(Exception ex)
        {
            return true;
        }


        public ITypeSerializationInfo GetSerializationInfo(Type type, bool createIfNecessary)
        {
            ITypeSerializationInfo info;
            if (!types.TryGetValue(type, out info))
            {
                if (createIfNecessary)
                {
                    AddType(type);
                    info = types[type];
                }
                else
                {
                    foreach (ITypeSerializationInfo tmp in types.Values)
                    {
                        if (tmp.Type.IsAssignableFrom(type))
                        {
                            if (info == null || info.Type.IsAssignableFrom(tmp.Type))
                            {
                                info = tmp;
                            }
                        }
                    }
                    types.Add(type, info);
                }
            } 
            return info;
        }

        public object Deserialize(TextReader input, bool fragment)
        {
            return Deserialize(input);
        }
    }


}