﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.ComponentModel;
using NMF.Transformations.Core.Properties;

namespace NMF.Transformations.Core
{
    /// <summary>
    /// This class represents a single computation within the transformation engine of NMF.Transformations
    /// </summary>
    /// <remarks>This class is visible to public as it provides reaction to delayness</remarks>
    public abstract class Computation : IPersistor, ITraceEntry
    {
        /// <summary>
        /// Creates a new computation for the given transformation rule with the given input arguments
        /// </summary>
        /// <param name="transformationRule">The transformation rule responsible for the transformation of the input data</param>
        /// <param name="context">The transformation context, in which the computation is done</param>
        protected Computation(GeneralTransformationRule transformationRule, ITransformationContext context)
        {
            if (transformationRule == null) throw new ArgumentNullException("transformationRule");

            TransformationRule = transformationRule;
            Context = context;
            MinOutputDelayLevel = transformationRule.OutputDelayLevel;
            MinTransformDelayLevel = transformationRule.TransformationDelayLevel;
        }

        /// <summary>
        /// Gets the transformation rule, which has been used to compute this computation
        /// </summary>
        public GeneralTransformationRule TransformationRule { get; private set; }

        /// <summary>
        /// Gets the context, in which the computation has been made
        /// </summary>
        public ITransformationContext Context { get; private set; }

        /// <summary>
        /// Gets the number of input arguments
        /// </summary>
        public int InputArguments { get { return TransformationRule.InputType.Length; } }

        /// <summary>
        /// Gets the input argument for this computation with the given index
        /// </summary>
        /// <param name="index">The index of the input parameter</param>
        /// <returns>The input parameter at the ith position</returns>
        public abstract object GetInput(int index);

        /// <summary>
        /// Copies the input of this computation into an array
        /// </summary>
        /// <returns>An array with the inputs</returns>
        public object[] CreateInputArray()
        {
            object[] inputs = new object[TransformationRule.InputType.Length];
            for (int i = 0; i < TransformationRule.InputType.Length; i++)
            {
                inputs[i] = GetInput(i);
            }
            return inputs;
        }
        
        /// <summary>
        /// Gets the output of this computation
        /// </summary>
        /// <exception cref="InvalidOperationException">This property may throw a DelayedOutputCreationException in case that the output has been tried to access, although the output creation was delayed</exception>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations")]
        public object Output
        {
            get
            {
                if (IsDelayed) ThrowDelayedException();
                return OutputCore;
            }
        }

        /// <summary>
        /// Throws an exception that the output has been accessed although the output of teh computation is delayed
        /// </summary>
        protected static void ThrowDelayedException()
        {
            throw new InvalidOperationException(Resources.ErrComputationOutputDelayed);
        }

        /// <summary>
        /// Marks that this computations requires another to be transformed.
        /// </summary>
        /// <param name="other">The other computation</param>
        /// <param name="executeBefore">A value indicating whether the other computation must be execute before or after the current computation</param>
        /// <remarks>The default implementation does nothing, so feel free to override. This method is intended to be called by NMF.Transformations, only.</remarks>
        protected virtual void MarkRequire(Computation other, bool executeBefore) { }

        internal void MarkRequireInternal(Computation other, bool executeBefore)
        {
            if (other != null)
            {
                if (executeBefore)
                {
                    DelayTransformationAtLeast(other.MinTransformDelayLevel);
                }
                else
                {
                    other.DelayTransformationAtLeast(MinTransformDelayLevel);
                }
                MarkRequire(other, executeBefore);
            }
        }

        internal OutputDelay OutputDelay { get; private set; }

        /// <summary>
        /// Gets or sets the output in a derived class
        /// </summary>
        protected abstract object OutputCore { get; set; }

        internal byte MinOutputDelayLevel { get; private set; }
        internal byte MinTransformDelayLevel { get; private set; }

        /// <summary>
        /// Configures the computation to create its output at minimum with the given delay level
        /// </summary>
        /// <param name="delayLevel">The minimum delay level for this computation</param>
        public void DelayOutputAtLeast(byte delayLevel)
        {
            MinOutputDelayLevel = Math.Max(MinOutputDelayLevel, delayLevel);
        }

        /// <summary>
        /// Configures the computation to be transformed at least with the given delay level
        /// </summary>
        /// <param name="delayLevel"></param>
        public void DelayTransformationAtLeast(byte delayLevel)
        {
            MinTransformDelayLevel = Math.Max(MinTransformDelayLevel, delayLevel);
        }



        /// <summary>
        /// Gets a value indicating whether the output creation for this computation is delayed
        /// </summary>
        public bool IsDelayed { get { return OutputDelay != null; } }


        /// <summary>
        /// Gets the output delay level
        /// </summary>
        internal byte OutputDelayLevel
        {
            get
            {
                if (OutputDelay == null)
                {
                    return 0;
                }
                else
                {
                    return OutputDelay.DelayLevel;
                }
            }
        }

        /// <summary>
        /// This event is fired as soon as the output of this computation is initialized
        /// </summary>
        public event EventHandler OutputInitialized;

        /// <summary>
        /// This event is fired as soon as the computation has been computed,
        /// i.e., the computation has been processed in the computation list
        /// </summary>
        public event EventHandler Computed;


        /// <summary>
        /// Fires the <see cref="OutputInitialized"/>-event
        /// </summary>
        /// <param name="e">The event arguments</param>
        protected virtual void OnOutputInitialized(EventArgs e)
        {
            if (OutputInitialized != null) OutputInitialized(this, e);
        }

        /// <summary>
        /// Fires the <see cref="Computed"/>-event
        /// </summary>
        /// <param name="e">The event arguments</param>
        protected virtual void OnComputed(EventArgs e)
        {
            if (Computed != null) Computed(this, e);
        }

        /// <summary>
        /// Initializes the transformation output
        /// </summary>
        /// <remarks>At this point, all the transformation outputs are created (also the delayed ones), thus, the trace is fully reliable</remarks>
        public abstract void Transform();

        /// <summary>
        /// Creates the output of this transformation rule
        /// </summary>
        /// <returns>The output for this transformation under this input</returns>
        /// <remarks>At this point, not all of the computations have created their output and thus, the trace is not complete. Use the OutputDelayLevel-feature to have the trace contain all elements created in earlier levels</remarks>
        public abstract object CreateOutput();

        /// <summary>
        /// Initializes the output for the actual computation
        /// </summary>
        /// <param name="output">The intended output</param>
        internal void InitializeOutput(object output)
        {
            OutputCore = output;
            if (OutputDelay != null)
            {
                OutputDelay.ExecutePersistors(output);
            }
            OutputDelay = null;
            OnOutputInitialized(EventArgs.Empty);
        }

        /// <summary>
        /// Delays the ouput of this computation with the given output delay
        /// </summary>
        /// <param name="delay"></param>
        internal void DelayOutput(OutputDelay delay)
        {
            if (delay == null) throw new ArgumentNullException("delay");

            if (OutputDelay != null)
            {
                delay.Persistors.AddRange(OutputDelay.Persistors);
            }
            OutputDelay = delay;
        }

        void IPersistor.Persist(object output)
        {
            InitializeOutput(output);
        }
    }
}
