﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace NMF.Transformations.Core
{
    /// <summary>
    /// This class is used to represent dependencies on more than one child object
    /// </summary>
    [DebuggerDisplay("{Representation}")]
    public class MultipleDependency : Dependency
    {
        internal Func<Computation, IEnumerable<object[]>> Selector { get; set; }
        internal Action<object, IEnumerable> Persistor { get; set; }

        internal string Representation
        {
            get
            {
                if (ExecuteBefore)
                {
                    return "Require many " + DependencyTransformation.ToString();
                }
                else
                {
                    return "Call many " + DependencyTransformation.ToString();
                }
            }
        }

        /// <summary>
        /// Calls the transformation dependency for the given computation
        /// </summary>
        /// <param name="computation">The computation that this dependency is to be called</param>
        public override void HandleDependency(Computation computation)
        {
            if (computation == null) throw new ArgumentNullException("computation");

            if (computation.IsDelayed && NeedOutput)
            {
                computation.OutputInitialized += CallComputation;
                return;
            }

            if (Filter != null && !Filter(computation)) return;
            var context = computation.Context;
            if (Selector != null)
            {
                if (computation.IsDelayed)
                {
                    var delay = computation.OutputDelay;
                    if (!NeedOutput)
                    {
                        HandleDelayedComputation(computation, context, delay);
                    }
                }
                else
                {
                    HandleNonDelayedComputation(computation, context);
                }
            }
        }

        private void HandleNonDelayedComputation(Computation computation, ITransformationContext context)
        {
            var output = computation.Output;
            var inCollection = Selector(computation);
            if (inCollection != null)
            {
                if (Persistor != null)
                {
                    Type listType = (typeof(List<>)).MakeGenericType(DependencyTransformation.OutputType);
                    IList list = System.Activator.CreateInstance(listType) as IList;

                    MultipleResultAwaitingPersistor delayPersistor = null;

                    foreach (var dependencyInput in inCollection)
                    {
                        GeneralTransformationRule dependent = DependencyTransformation;
                        var comp2 = context.CallTransformation(dependent, dependencyInput);
                        if (!comp2.IsDelayed)
                        {
                            list.Add(comp2.Output);
                        }
                        else
                        {
                            if (delayPersistor == null)
                            {
                                delayPersistor = new MultipleResultAwaitingPersistor()
                                {
                                    List = list,
                                    Persistor = Persistor,
                                    Target = output
                                };
                            }
                            delayPersistor.WaitFor(comp2);
                        }

                        computation.MarkRequireInternal(comp2, ExecuteBefore);

                    }
                    if (delayPersistor == null) Persistor(output, list);
                }
                else
                {
                    foreach (var dependencyInput in inCollection)
                    {
                        GeneralTransformationRule dependent = DependencyTransformation;
                        var comp2 = context.CallTransformation(dependent, dependencyInput);

                        computation.MarkRequireInternal(comp2, ExecuteBefore);
                    }
                }
            }
        }

        private void HandleDelayedComputation(Computation computation, ITransformationContext context, OutputDelay delay)
        {
            var inCollection = Selector(computation);
            if (inCollection != null)
            {
                if (Persistor != null)
                {
                    Type listType = (typeof(List<>)).MakeGenericType(DependencyTransformation.OutputType);
                    IList list = System.Activator.CreateInstance(listType) as IList;

                    MultipleResultAwaitingPersistor delayPersistor = null;

                    foreach (var dependencyInput in inCollection)
                    {
                        GeneralTransformationRule dependent = DependencyTransformation;
                        var comp2 = context.CallTransformation(dependent, dependencyInput);
                        if (!comp2.IsDelayed)
                        {
                            list.Add(comp2.Output);
                        }
                        else
                        {
                            computation.DelayOutputAtLeast(comp2.MinOutputDelayLevel);
                            if (delayPersistor == null)
                            {
                                delayPersistor = new MultipleResultAwaitingPersistor()
                                {
                                    List = list,
                                    Persistor = Persistor
                                };
                            }
                            delayPersistor.WaitFor(comp2);
                        }

                        if (ExecuteBefore)
                        {
                            computation.DelayTransformationAtLeast(comp2.MinTransformDelayLevel);
                        }
                        else
                        {
                            comp2.DelayTransformationAtLeast(computation.MinTransformDelayLevel);
                        }
                    }
                    if (delayPersistor == null)
                    {
                        delay.Persistors.Add(new MultipleItemPersistor()
                        {
                            List = list,
                            Persistor = Persistor
                        });
                    }
                    else
                    {
                        delay.Persistors.Add(delayPersistor);
                    }
                }
                else
                {
                    foreach (var dependencyInput in inCollection)
                    {
                        GeneralTransformationRule dependent = DependencyTransformation;
                        var comp2 = context.CallTransformation(dependent, dependencyInput);

                        computation.MarkRequireInternal(comp2, ExecuteBefore);
                    }
                }
            }
        }

        private void CallComputation(object sender, EventArgs e)
        {
            Computation c = sender as Computation;
            HandleDependency(c);
            c.OutputInitialized -= CallComputation;
        }
    }

    
}
