﻿using NMF.Utilities;
using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Dynamic;

namespace NMF.Transformations.Core
{
    /// <summary>
    /// This is the most important class of NMF.Transformations as it handles all the transformation
    /// </summary>
    public class TransformationContext : ITransformationContext
    {
        private Dictionary<object[], List<ITraceEntry>> computationsMade = new Dictionary<object[], List<ITraceEntry>>(ItemEqualityComparer<object>.Instance);
        private Dictionary<GeneralTransformationRule, List<ITraceEntry>> computationsByTransformationRule = new Dictionary<GeneralTransformationRule, List<ITraceEntry>>();
        private List<ICollection<Computation>> computationOrder;
        private List<ICollection<DelayedOutputCreation>> delay;

        private List<object[]> inputs = new List<object[]>();
        private List<object> outputs = new List<object>();

        /// <summary>
        /// Gets the parent transformation, that the context is based upon
        /// </summary>
        public Transformation Transformation { get; private set; }
        private byte currentTransformationDelay = 0;
        private byte currentOutputDelay = 0;
        private ITransformationTrace trace;

        /// <summary>
        /// Creates a new transformation context for the given transformation
        /// </summary>
        /// <param name="transformation">The transformation, a context should be generated for</param>
        public TransformationContext(Transformation transformation)
        {
            if (transformation == null) throw new ArgumentNullException("transformation");
            Transformation = transformation;
            trace = new TransformationContextTrace(this);
        }

        #region CallTransformation & Outsourced functions

        /// <summary>
        /// Calls the given transformation with the specified input
        /// </summary>
        /// <param name="input">The input for the transformation rule</param>
        /// <param name="transformationRule">The rule that should be applied</param>
        /// <returns>The computation that handles this request</returns>
        public Computation CallTransformation(GeneralTransformationRule transformationRule, params object[] input)
        {
            if (transformationRule == null) throw new ArgumentNullException("transformationRule");
            if (input == null) throw new ArgumentNullException("input");

            List<ITraceEntry> computations;
            if (!computationsMade.TryGetValue(input, out computations))
            {
                computations = new List<ITraceEntry>();
                computationsMade.Add(input, computations);
            }

            var comp = computations.Where(cpt => cpt.TransformationRule == transformationRule).OfType<Computation>().FirstOrDefault();
            if (comp == null)
            {
                comp = transformationRule.CreateComputation(input, this);

                if (comp == null) return null;

                comp.DelayOutput(new OutputDelay());

                computations.Add(comp);
                AddTraceEntry(comp);

                CallDependencies(comp, true);

                // The transformation output is only generated when we are handling the base transformation rule,
                // because this is always required
                if (comp.IsDelayed)
                {
                    Stack<Computation> dependantComputes = new Stack<Computation>();
                    var ruleStack = Transformation.ComputeInstantiatingTransformationRulePath(comp);
                    var delayLevel = comp.MinOutputDelayLevel;

                    var computes = new List<Computation>();

                    while (ruleStack.Count > 0)
                    {
                        var rule = ruleStack.Pop();
                        var comp2 = computations.Where(cmp => cmp.TransformationRule == rule).OfType<Computation>().FirstOrDefault();
                        if (comp2 == null)
                        {
                            comp2 = rule.CreateComputation(input, this);
                            computations.Add(comp2);
                            AddTraceEntry(comp2);
                            comp2.DelayOutput(comp.OutputDelay);
                            dependantComputes.Push(comp2);
                        }

                        // in case comp2 is not yet handled, a delay does not yet exist and thus
                        // DelayLevel < minDelayLevel
                        delayLevel = Math.Max(delayLevel, Math.Max(comp2.OutputDelayLevel, comp2.MinOutputDelayLevel));
                        computes.Add(comp2);
                    }

                    // delay the call of dependencies
                    // this prevents the issue arising from computations calling their parents that come later in the stack
                    foreach (var comp2 in dependantComputes)
                    {
                        CallDependencies(comp2, true);
                    }

                    delayLevel += currentOutputDelay;

                    if (delayLevel == 0)
                    {
                        var createRule = computes[0];

                        // Generate the output
                        var output = createRule.CreateOutput();

                        for (int i = computes.Count - 1; i >= 0; i--)
                        {
                            computes[i].InitializeOutput(output);
                        }
                    }
                    else
                    {
                        //Save computations into Delay
                        Delay(delayLevel, computes);
                    }


                    for (int i = computes.Count - 1; i >= 0; i--)
                    {
                        AddToComputationOrder(computes[i], currentTransformationDelay);
                    }

                    for (int i = computes.Count - 1; i >= 0; i--)
                    {
                        CallDependencies(computes[i], false);
                    }
                }
            }
            return comp;
        }

        /// <summary>
        /// Creates a trace entry for the given computation object
        /// </summary>
        /// <remarks>Override for custom trace entries. A null-check for the argument is not required.</remarks>
        /// <param name="computation">The computation that needs to be added to the trace</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1062:Validate arguments of public methods", MessageId = "0")]
        protected virtual void AddTraceEntry(Computation computation)
        {
            List<ITraceEntry> comps;
            if (!computationsByTransformationRule.TryGetValue(computation.TransformationRule, out comps))
            {
                comps = new List<ITraceEntry>();
                computationsByTransformationRule.Add(computation.TransformationRule, comps);
            }
            comps.Add(computation);
        }

        private void AddToComputationOrder(Computation c, byte level)
        {
            level = Math.Max(level, c.MinTransformDelayLevel);
            ICollection<Computation> list;
            if (computationOrder == null) computationOrder = new List<ICollection<Computation>>();
            if (computationOrder.Count > level)
            {
                list = computationOrder[level];
                if (list == null)
                {
                    list = new List<Computation>();
                    computationOrder[level] = list;
                }
            }
            else
            {
                while (computationOrder.Count < level)
                {
                    computationOrder.Add(null);
                }
                list = new List<Computation>();
                computationOrder.Add(list);
            }
            list.Add(c);
        }

        private void Delay(byte delayLevel, List<Computation> computes)
        {
            delayLevel = Math.Max(delayLevel, currentOutputDelay);
            ICollection<DelayedOutputCreation> list;
            if (delay == null) delay = new List<ICollection<DelayedOutputCreation>>();
            if (delay.Count > delayLevel)
            {
                list = delay[delayLevel];
                if (list == null)
                {
                    list = new List<DelayedOutputCreation>();
                    delay[delayLevel] = list;
                }
            }
            else
            {
                while (delay.Count < delayLevel)
                {
                    delay.Add(null);
                }
                list = new List<DelayedOutputCreation>();
                delay.Add(list);
            }
            var output = new DelayedOutputCreation(computes);
            list.Add(output);
        }

        private static void CallDependencies(Computation c, bool executeBefore)
        {
            foreach (var requirement in c.TransformationRule.Dependencies)
            {
                if (requirement.ExecuteBefore == executeBefore)
                {
                    requirement.HandleDependency(c);
                }
            }
        }

        /// <summary>
        /// Executes all computations registered,but not already handled
        /// </summary>
        public void ExecuteAllTransformations()
        {
            if (computationOrder != null)
            {
                currentOutputDelay = 0;
                while (currentTransformationDelay < computationOrder.Count)
                {
                    //first increment currentTransformationDelay to avoid conflicts
                    //because items are put in the collection while there is an Enumerator
                    //on the collection
                    currentTransformationDelay++;
                    var collection = computationOrder[currentTransformationDelay - 1];
                    if (collection != null)
                    {
                        foreach (var item in collection)
                        {
                                item.Transform();
                            }
                    }
                    CreateDelayedOutputs();
                }
            }
        }

        /// <summary>
        /// Creates the outputs of all delayed computations
        /// </summary>
        public void CreateDelayedOutputs()
        {
            if (delay != null)
            {
                while (currentOutputDelay < delay.Count)
                {
                    currentOutputDelay++;
                    var delayLevel = delay[currentOutputDelay - 1];
                    if (delayLevel != null)
                    {
                        foreach (var item in delayLevel)
                        {
                                item.CreateDelayedOutput();
                            }
                        delayLevel.Clear();
                    }
                }
                delay.Clear();
                currentOutputDelay = 0;
            }
        }

        #endregion

        #region Trace Helpers

        /// <summary>
        /// Gets all rules that apply the given signature
        /// </summary>
        /// <param name="input">The input argument type list</param>
        /// <param name="output">The output type</param>
        /// <returns>A collection with all the rules that have the given signature</returns>
        public IEnumerable<GeneralTransformationRule> GetRules(Type[] input, Type output)
        {
            return Transformation.GetRulesForTypeSignature(input, output);
        }

        /// <summary>
        /// Gets any rules that apply the given signature
        /// </summary>
        /// <param name="input">The input argument type list</param>
        /// <param name="output">The output type</param>
        /// <returns>A random rule that has the given signature</returns>
        public GeneralTransformationRule GetRule(Type[] input, Type output)
        {
            return Transformation.GetRuleForTypeSignature(input, output);
        }

        #endregion

        #region Bag & Data

        private dynamic bag = new ExpandoObject();
        private IDictionary<object, object> data = new Dictionary<object, object>();

        /// <summary>
        /// Gets a Bag, where dynamic data can be added
        /// </summary>
        /// <remarks>The value of this property is an ExpandoObject, so that the bag can be easily extended with new properties</remarks>
        public dynamic Bag
        {
            get { return bag; }
        }


        /// <summary>
        /// Gets a data dictionary, where data set during the transformation can be added
        /// </summary>
        public IDictionary<object, object> Data
        {
            get { return data; }
        }

        #endregion

        #region Trace

        /// <summary>
        /// Represents the trace class for TransformationContext
        /// </summary>
        protected class TransformationContextTrace : AbstractTrace, ITransformationTrace
        {
            private Dictionary<object[], List<ITraceEntry>> computationsMade;
            private Dictionary<GeneralTransformationRule, List<ITraceEntry>> computationsByTransformationRule;
            private TransformationContext context;

            /// <summary>
            /// Creates a new trace class for the given TraceContext
            /// </summary>
            /// <param name="context">The trace class for which the trace should be generated</param>
            public TransformationContextTrace(TransformationContext context)
            {
                if (context == null) throw new ArgumentNullException("context");
                this.computationsMade = context.computationsMade;
                this.computationsByTransformationRule = context.computationsByTransformationRule;
                this.context = context;
            }

            /// <summary>
            /// Gets a collection of the underlying computations
            /// </summary>
            public override IEnumerable<ITraceEntry> Computations
            {
                get { return context.Computations
                    .Except(revoked)
                    .Concat(published); }
            }

            #region Trace accelerated by "computationsMade"

            /// <summary>
            /// Traces the computations based upon the specified input
            /// </summary>
            /// <returns>The computations with the given inputs</returns>
            /// <param name="input">The input arguments</param>
            public override IEnumerable<ITraceEntry> Trace(object[] input)
            {
                List<ITraceEntry> comps;
                if (computationsMade.TryGetValue(input, out comps))
                {
                    return comps.AsReadOnly();
                }
                return Enumerable.Empty<ITraceEntry>();
            }

            /// <summary>
            /// Traces the computations of the specified inputs with the specified transformation rules
            /// </summary>
            /// <param name="rule">The transformation rules that transformed the specified inputs</param>
            /// <param name="inputs">A collection of input arguments</param>
            /// <returns>A collection of computations</returns>
            public override IEnumerable<ITraceEntry> TraceManyIn(GeneralTransformationRule rule, IEnumerable<object[]> inputs)
            {
                if (rule == null) throw new ArgumentNullException("rule");
                if (inputs == null) return Enumerable.Empty<ITraceEntry>();
                List<ITraceEntry> result = new List<ITraceEntry>();
                foreach (var input in inputs)
                {
                    List<ITraceEntry> comps;
                    if (computationsMade.TryGetValue(input, out comps))
                    {
                        result.AddRange(comps.Where(c => c.TransformationRule == rule));
                    }
                }
                return result;
            }

            /// <summary>
            /// Traces the computation based upon the specified input with the specified transformation rule
            /// </summary>
            /// <param name="rule">The transformation rule the object was transformed with</param>
            /// <returns>The computation or null, if there was none</returns>
            /// <param name="input">The input arguments</param>
            public override IEnumerable<ITraceEntry> TraceIn(GeneralTransformationRule rule, params object[] input)
            {
                if (rule == null) throw new ArgumentNullException("rule");
                List<ITraceEntry> comps;
                if (computationsMade.TryGetValue(input, out comps))
                {
                    return comps.Where(c => c.TransformationRule == rule);
                }
                return Enumerable.Empty<ITraceEntry>();
            }

            /// <summary>
            /// Traces the computations of the specified inputs that match the given type signature
            /// </summary>
            /// <param name="inputs">A collection of input arguments</param>
            /// <param name="inputTypes">The input types</param>
            /// <param name="outputType">The output types</param>
            /// <returns>A collection of computations</returns>
            public override IEnumerable<ITraceEntry> TraceMany(Type[] inputTypes, Type outputType, IEnumerable<object[]> inputs)
            {
                if (inputs == null) return Enumerable.Empty<ITraceEntry>();
                List<ITraceEntry> result = new List<ITraceEntry>();
                foreach (var input in inputs)
                {
                    List<ITraceEntry> comps;
                    if (computationsMade.TryGetValue(input, out comps))
                    {
                        result.AddRange(comps);
                    }
                }
                return result;
            }

            #endregion

            #region Trace accelerated by "computationsByTransformationRule"

            /// <summary>
            /// Traces all computations with any inputs that math the given filters with the specified transformation rule
            /// </summary>
            /// <param name="rule">The transformation rule</param>
            /// <returns>A collection with all computations made under these circumstances</returns>
            public override IEnumerable<ITraceEntry> TraceAllIn(GeneralTransformationRule rule)
            {
                if (rule == null) throw new ArgumentNullException("rule");
                List<ITraceEntry> comps;
                if (computationsByTransformationRule.TryGetValue(rule, out comps))
                {
                    return comps.AsReadOnly();
                }
                else
                {
                    return Enumerable.Empty<Computation>();
                }
            }

            #endregion

            private HashSet<ITraceEntry> revoked = new HashSet<ITraceEntry>();
            private HashSet<ITraceEntry> published = new HashSet<ITraceEntry>();

            /// <summary>
            /// Revokes the given computation and deletes it from the trace
            /// </summary>
            /// <param name="traceEntry">The computation that is to be revoked</param>
            public override void RevokeEntry(ITraceEntry traceEntry)
            {
                if (traceEntry == null) throw new ArgumentNullException("traceEntry");

                if (published.Contains(traceEntry))
                {
                    published.Remove(traceEntry);
                }
                else
                {
                    revoked.Add(traceEntry);
                }

                List<ITraceEntry> comps;
                if (computationsMade.TryGetValue(traceEntry.CreateInputArray(), out comps))
                {
                    comps.Remove(traceEntry);
                }
                if (computationsByTransformationRule.TryGetValue(traceEntry.TransformationRule, out comps))
                {
                    comps.Remove(traceEntry);
                }
            }

            /// <summary>
            /// Publishes the given computation to the trace
            /// </summary>
            /// <param name="traceEntry">The computation that should be added to the trace</param>
            public override void PublishEntry(ITraceEntry traceEntry)
            {
                if (traceEntry == null) throw new ArgumentNullException("traceEntry");

                if (revoked.Contains(traceEntry))
                {
                    revoked.Remove(traceEntry);
                }
                else
                {
                    var success = published.Add(traceEntry);
                    if (!success) return;
                }

                var box = traceEntry.CreateInputArray();
                List<ITraceEntry> comps;
                if (!computationsMade.TryGetValue(box, out comps))
                {
                    comps = new List<ITraceEntry>();
                    computationsMade.Add(box, comps);
                }
                if (!comps.Contains(traceEntry))
                {
                    comps.Add(traceEntry);
                    if (!computationsByTransformationRule.TryGetValue(traceEntry.TransformationRule, out comps))
                    {
                        comps = new List<ITraceEntry>();
                        computationsByTransformationRule.Add(traceEntry.TransformationRule, comps);
                    }
                    comps.Add(traceEntry);
                }
                else
                {
                    published.Remove(traceEntry);
                }
            }
        }

        #endregion

        /// <summary>
        /// Calls the transformation context to finish any things yet undone
        /// </summary>
        public void FinishTransformation()
        {
            CreateDelayedOutputs();
            ExecuteAllTransformations();
        }

        /// <summary>
        /// Gets all computations (for custom trace purposes)
        /// </summary>
        public IEnumerable<Computation> Computations
        {
            get { return computationOrder.SelectMany(o => o); }
        }

        /// <summary>
        /// Gets the object responsible for trace operations for this transformation context
        /// </summary>
        public ITransformationTrace Trace
        {
            get { return trace; }
        }

        /// <summary>
        /// Gets the input of the transformation context
        /// </summary>
        /// <remarks>If the transformation has multiple inputs, this returns the first input</remarks>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
        public object[] Input
        {
            get { return inputs.FirstOrDefault(); }
        }

        /// <summary>
        /// Gets a collection of inputs
        /// </summary>
        public IList<object[]> Inputs
        {
            get { return inputs; }
        }

        /// <summary>
        /// Gets the output of the transformation context
        /// </summary>
        /// <remarks>If the transformation has multiple outputs, this property returns the first output</remarks>
        public object Output
        {
            get { return outputs.FirstOrDefault(); }
        }

        /// <summary>
        /// Gets a collection of outputs
        /// </summary>
        public IList<object> Outputs
        {
            get { return outputs; }
        }
    }
}
