﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NMF.Transformations.Linq;

using System.Collections.Generic;

using NMF.Tests;

namespace NMF.Transformations.Tests.UnitTests.Relational
{
    [TestClass]
    public class RelationalTest
    {
        private GrowableSource<string> source = new GrowableSource<string>();
        private Func<int, IGrowableSource<string>> sourceFunc;

        [TestInitialize]
        public void TestInit()
        {
            sourceFunc = i => source;
        }

        [TestMethod]
        public void Transformations_Relational_Where1()
        {
            bool filterCalled = false;
            bool filterResult = true;
            bool nextItemCalled = false;
            var filtered = source.Where(s => { filterCalled = true; return filterResult; });
            filtered.NextItem += item =>
            {
                Assert.AreEqual("a", item);
                nextItemCalled = true;
            };

            filtered.AssertEmpty();
            Assert.IsFalse(filterCalled);

            source.Add("a");

            Assert.IsTrue(filterCalled);
            Assert.IsTrue(nextItemCalled);
            filtered.AssertContainsOnly("a");

            nextItemCalled = false;
            filterCalled = false;
            filterResult = false;

            source.Add("b");

            Assert.IsTrue(filterCalled);
            Assert.IsFalse(nextItemCalled);

            filterCalled = false;
            filterResult = true;

            filtered.Dispose();
            source.Add("c");

            Assert.IsFalse(filterCalled);
            Assert.IsFalse(nextItemCalled);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Where_Exception1()
        {
            var result = source.Where(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Where_Exception2()
        {
            var result = (null as IGrowableSource<string>).Where(s => true);
        }

        [TestMethod]
        public void Transformations_Relational_Select1()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;
            int expectedResult = 1;

            var select = source.Select(s => { selectorCalled = true; return s.Length; });
            select.NextItem += item =>
            {
                Assert.AreEqual(expectedResult, item);
                nextItemCalled = true;
            };

            select.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            select.AssertContainsOnly(1);

            selectorCalled = false;
            nextItemCalled = false;
            expectedResult = 2;

            source.Add("aa");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            select.AssertContainsOnly(1, 2);

            selectorCalled = false;
            nextItemCalled = false;
            expectedResult = 0;

            select.Dispose();

            source.Add("");

            Assert.IsFalse(selectorCalled);
            Assert.IsFalse(nextItemCalled);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Select_Exception1()
        {
            var result = source.Select(null as Func<string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Select_Exception2()
        {
            var result = (null as IGrowableSource<string>).Select(s => new GenericParameterHelper());
        }

        [TestMethod]
        public void Transformations_Relational_AsGrowable()
        {
            var list = new List<string>();

            var result = list.AsGrowable();

            result.AssertNotNull();
            Assert.IsTrue(result is GrowableSource<string>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_AsGrowable_Exception()
        {
            var result = (null as ICollection<GenericParameterHelper>).AsGrowable();
        }

        [TestMethod]
        public void Transformations_Relational_SelectMany1()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;

            var result = source.SelectMany(s => source, (s1, s2) => { selectorCalled = true; return Tuple.Create(s1, s2); });
            result.NextItem += item =>
            {
                nextItemCalled = true;
            };

            result.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "a"));

            selectorCalled = false;
            nextItemCalled = false;

            source.Add("b");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "a"), Tuple.Create("a", "b"), Tuple.Create("b", "a"), Tuple.Create("b", "b"));

            selectorCalled = false;
            nextItemCalled = false;

            result.Dispose();
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception1()
        {
            var result = source.SelectMany(s => source, null as Func<string, string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception2()
        {
            var result = source.SelectMany(null as Func<string, IGrowableSource<GenericParameterHelper>>, (s, g) => new GenericParameterHelper());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception3()
        {
            var result = (null as IGrowableSource<GenericParameterHelper>).SelectMany(s => source, (g, s) => new GenericParameterHelper());
        }

        [TestMethod]
        public void Transformations_Relational_SelectMany2()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;

            var list = new List<string>() { "d" };

            var result = source.SelectMany(s => list, (s1, s2) => { selectorCalled = true; return Tuple.Create(s1, s2); });
            result.NextItem += item =>
            {
                nextItemCalled = true;
            };

            result.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "d"));

            selectorCalled = false;
            nextItemCalled = false;

            source.Add("b");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "d"), Tuple.Create("b", "d"));

            selectorCalled = false;
            nextItemCalled = false;

            result.Dispose();
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception4()
        {
            var result = source.SelectMany(s => new List<string>() {"d"}, null as Func<string, string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception5()
        {
            var result = source.SelectMany(null as Func<string, IEnumerable<GenericParameterHelper>>, (s, g) => new GenericParameterHelper());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception6()
        {
            var result = (null as IGrowableSource<GenericParameterHelper>).SelectMany(s => new List<string>() { "d" }, (g, s) => new GenericParameterHelper());
        }


        [TestMethod]
        public void Transformations_Relational_Where2()
        {
            bool filterCalled = false;
            bool filterResult = true;
            bool nextItemCalled = false;
            var filtered = (sourceFunc.Where(s => { filterCalled = true; return filterResult; }))(42);
            filtered.NextItem += item =>
            {
                Assert.AreEqual("a", item);
                nextItemCalled = true;
            };

            filtered.AssertEmpty();
            Assert.IsFalse(filterCalled);

            source.Add("a");

            Assert.IsTrue(filterCalled);
            Assert.IsTrue(nextItemCalled);
            filtered.AssertContainsOnly("a");

            nextItemCalled = false;
            filterCalled = false;
            filterResult = false;

            source.Add("b");

            Assert.IsTrue(filterCalled);
            Assert.IsFalse(nextItemCalled);

            filterCalled = false;
            filterResult = true;

            filtered.Dispose();
            source.Add("c");

            Assert.IsFalse(filterCalled);
            Assert.IsFalse(nextItemCalled);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Where_Exception3()
        {
            var result = sourceFunc.Where(null);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Where_Exception4()
        {
            var result = (null as Func<GenericParameterHelper,IGrowableSource<GenericParameterHelper>>).Where(s => true);
        }

        [TestMethod]
        public void Transformations_Relational_Select2()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;
            int expectedResult = 1;

            var select = sourceFunc.Select(s => { selectorCalled = true; return s.Length; })(42);
            select.NextItem += item =>
            {
                Assert.AreEqual(expectedResult, item);
                nextItemCalled = true;
            };

            select.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            select.AssertContainsOnly(1);

            selectorCalled = false;
            nextItemCalled = false;
            expectedResult = 2;

            source.Add("aa");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            select.AssertContainsOnly(1, 2);

            selectorCalled = false;
            nextItemCalled = false;
            expectedResult = 0;

            select.Dispose();

            source.Add("");

            Assert.IsFalse(selectorCalled);
            Assert.IsFalse(nextItemCalled);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Select_Exception3()
        {
            var result = source.Select(null as Func<string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_Select_Exception4()
        {
            var result = (null as IGrowableSource<string>).Select(s => new GenericParameterHelper());
        }

        [TestMethod]
        public void Transformations_Relational_SelectMany3()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;

            var result = sourceFunc.SelectMany(s => sourceFunc, (s1, s2) => { selectorCalled = true; return Tuple.Create(s1, s2); })(42);
            result.NextItem += item =>
            {
                nextItemCalled = true;
            };

            result.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "a"));

            selectorCalled = false;
            nextItemCalled = false;

            source.Add("b");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "a"), Tuple.Create("a", "b"), Tuple.Create("b", "a"), Tuple.Create("b", "b"));

            selectorCalled = false;
            nextItemCalled = false;

            result.Dispose();
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception7()
        {
            var result = sourceFunc.SelectMany(s => sourceFunc, null as Func<string, string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception8()
        {
            var result = sourceFunc.SelectMany(null as Func<string, Func<int,IGrowableSource<GenericParameterHelper>>>, (s, g) => new GenericParameterHelper());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception9()
        {
            var result = (null as Func<int,IGrowableSource<GenericParameterHelper>>).SelectMany(s => sourceFunc, (g, s) => new GenericParameterHelper());
        }

        [TestMethod]
        public void Transformations_Relational_SelectMany4()
        {
            bool selectorCalled = false;
            bool nextItemCalled = false;

            var list = new List<string>() { "d" };

            var result = sourceFunc.SelectMany(s => new Func<int, IEnumerable<string>>(i => list), (s1, s2) => { selectorCalled = true; return Tuple.Create(s1, s2); })(42);
            result.NextItem += item =>
            {
                nextItemCalled = true;
            };

            result.AssertEmpty();

            source.Add("a");

            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "d"));

            selectorCalled = false;
            nextItemCalled = false;

            source.Add("b");
            Assert.IsTrue(nextItemCalled);
            Assert.IsTrue(selectorCalled);
            result.AssertContainsOnly(Tuple.Create("a", "d"), Tuple.Create("b", "d"));

            selectorCalled = false;
            nextItemCalled = false;

            result.Dispose();
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception10()
        {
            var result = source.SelectMany(s => new List<string>() { "d" }, null as Func<string, string, GenericParameterHelper>);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception11()
        {
            var result = source.SelectMany(null as Func<string, IEnumerable<GenericParameterHelper>>, (s, g) => new GenericParameterHelper());
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentNullException))]
        public void Transformations_Relational_SelectMany_Exception12()
        {
            var result = (null as IGrowableSource<GenericParameterHelper>).SelectMany(s => new List<string>() { "d" }, (g, s) => new GenericParameterHelper());
        }
    }
}
