﻿using NMF.Transformations.Core;
using NMF.Transformations.Properties;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;

namespace NMF.Transformations
{
    /// <summary>
    /// Defines a simple transformation rule of a transformation that has one input argument and an output
    /// </summary>
    /// <typeparam name="TIn">The type of the input argument</typeparam>
    /// <typeparam name="TOut">The type of the output</typeparam>
    /// <remarks>Simple means that the transformation rule does not require a custom computation class</remarks>
    public class TransformationRule<TIn, TOut> : TransformationRuleBase<TIn, TOut>
        where TIn : class
        where TOut : class
    {
        private sealed class SimpleComputation : TransformationComputation<TIn, TOut>
        {
            public SimpleComputation(TransformationRule<TIn, TOut> transformationRule, TIn input, ITransformationContext context)
                : base(transformationRule, context, input) { }

            public override void Transform()
            {
                if (Input != null)
                {
                    (TransformationRule as TransformationRule<TIn, TOut>).Transform(Input, Output, Context);
                    OnComputed(EventArgs.Empty);
                }
            }

            public override object CreateOutput()
            {
                if (Input != null)
                {
                    return (TransformationRule as TransformationRule<TIn, TOut>).CreateOutput(Input, Context);
                }
                else
                {
                    return null;
                }
            }
        }

        /// <summary>
        /// Creates a new Computation instance for this transformation rule or the given input 
        /// </summary>
        /// <param name="input">The input arguments for this computation</param>
        /// <param name="context">The context for this computation</param>
        /// <returns>A computation object</returns>
        public sealed override Computation CreateComputation(object[] input, ITransformationContext context)
        {
            if (input == null) return null;
            if (input.Length != 1) throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, Resources.ErrTransformationRuleWrongNumberOfArguments, this.GetType().Name));
            return new SimpleComputation(this, input[0] as TIn, context);
        }

        /// <summary>
        /// Creates the output of this transformation rule
        /// </summary>
        /// <param name="input">The input of the transformation rule</param>
        /// <param name="context">The context (and trace!) object</param>
        /// <returns>The output for this transformation under this input</returns>
        /// <remarks>At this point, not all of the computations have created their output and thus, the trace is not complete. Use the OutputDelayLevel-feature to have the trace contain all elements created in earlier levels</remarks>
        public virtual TOut CreateOutput(TIn input, ITransformationContext context)
        {
            return Activator.CreateInstance(OutputType) as TOut;
        }

        /// <summary>
        /// Initializes the transformation output
        /// </summary>
        /// <param name="input">The input of the transformation rule</param>
        /// <param name="output">The output of the transformation rule generated earlier</param>
        /// <param name="context">The context (and trace!) object</param>
        /// <remarks>At this point, all the transformation outputs are created (also the delayed ones), thus, the trace is fully reliable</remarks>
        public virtual void Transform(TIn input, TOut output, ITransformationContext context) { }
    }
}
