﻿using NMF.Transformations.Core;
using NMF.Transformations.Properties;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;

namespace NMF.Transformations
{

    /// <summary>
    /// Defines a simple transformation rule of a transformation that has one input argument and an output
    /// </summary>
    /// <typeparam name="TIn1">The type of the first input argument</typeparam>
    /// <typeparam name="TIn2">The type of the second input argument</typeparam>
    /// <typeparam name="TOut">The type of the output</typeparam>
    /// <remarks>Simple means that the transformation rule does not require a custom computation class</remarks>
    
    public class TransformationRule<TIn1, TIn2, TOut> : TransformationRuleBase<TIn1, TIn2, TOut>
        where TIn1 : class
        where TIn2 : class
        where TOut : class
    {
        private sealed class SimpleComputation : TransformationComputation<TIn1, TIn2, TOut>
        {
            public SimpleComputation(TransformationRule<TIn1, TIn2, TOut> transformationRule, TIn1 input1, TIn2 input2, ITransformationContext context)
                : base(transformationRule, context, input1, input2) { }

            public override void Transform()
            {
                (TransformationRule as TransformationRule<TIn1, TIn2, TOut>).Transform(Input1, Input2, Output, Context);
                OnComputed(EventArgs.Empty);
            }

            public override object CreateOutput()
            {
                return (TransformationRule as TransformationRule<TIn1, TIn2, TOut>).CreateOutput(Input1, Input2, Context);
            }
        }


        /// <summary>
        /// Creates a new Computation instance for this transformation rule or the given input 
        /// </summary>
        /// <param name="input">The input arguments for this computation</param>
        /// <param name="context">The context for this computation</param>
        /// <returns>A computation object</returns>
        public sealed override Computation CreateComputation(object[] input, ITransformationContext context)
        {
            if (input == null) return null;
            if (input.Length != 2) throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, Resources.ErrTransformationRuleWrongNumberOfArguments, this.GetType().Name));
            return new SimpleComputation(this, input[0] as TIn1, input[1] as TIn2, context);
        }



        /// <summary>
        /// Creates the output of this transformation rule
        /// </summary>
        /// <param name="input1">The first input of the transformation rule</param>
        /// <param name="input2">The second input of the transformation rule</param>
        /// <param name="context">The context (and trace!) object</param>
        /// <returns>The output for this transformation under this input</returns>
        /// <remarks>At this point, not all of the computations have created their output and thus, the trace is not complete. Use the OutputDelayLevel-feature to have the trace contain all elements created in earlier levels</remarks>
        public virtual TOut CreateOutput(TIn1 input1, TIn2 input2, ITransformationContext context)
        {
            return Activator.CreateInstance(OutputType) as TOut;
        }

        /// <summary>
        /// Initializes the transformation output
        /// </summary>
        /// <param name="input1">The first input of the transformation rule</param>
        /// <param name="input2">The second input of the transformation rule</param>
        /// <param name="output">The output of the transformation rule generated earlier</param>
        /// <param name="context">The context (and trace!) object</param>
        /// <remarks>At this point, all the transformation outputs are created (also the delayed ones), thus, the trace is fully reliable</remarks>
        public virtual void Transform(TIn1 input1, TIn2 input2, TOut output, ITransformationContext context) { }
        
    }
}
