﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace NMF.Transformations.Linq
{
    /// <summary>
    /// This class represents the default implementation to fulfill the C# bind-operator (SelectMany). 
    /// </summary>
    /// <typeparam name="T1">The inner type of the outer monad instance</typeparam>
    /// <typeparam name="T2">The inner type of the inner monad instance</typeparam>
    /// <typeparam name="TTarget">The inner type of the target monad instance</typeparam>
    /// <remarks>Please see the documentation of the SelectMany extension method and its use with the LINQ-syntax for more details</remarks>
    internal class CompositeSource<T1, T2, TTarget> : IGrowableSource<TTarget>
    {
        public CompositeSource(IGrowableSource<T1> source, Func<T1, IGrowableSource<T2>> func, Func<T1, T2, TTarget> selector)
        {
            Source = source;
            Func = func;
            Selector = selector;

            source.NextItem += OnNextSourceItem;
        }

        void OnNextSourceItem(T1 item)
        {
            var subSource = Func(item);
            foreach (var element in subSource)
            {
                OnNextItem(Selector(item, element));
            }
            subSource.NextItem += element => OnNextItem(Selector(item, element));
        }

        private IGrowableSource<T1> Source { get; set; }
        private Func<T1, IGrowableSource<T2>> Func { get; set; }
        private Func<T1, T2, TTarget> Selector { get; set; }

        protected virtual void OnNextItem(TTarget nextItem)
        {
            if (NextItem != null) NextItem(nextItem);
        }

        public event Action<TTarget> NextItem;

        public void Dispose()
        {
            foreach (var item in Source)
            {
                var sub = Func(item);
                if (sub != null) sub.Dispose();
            }
            Source.NextItem -= OnNextSourceItem;
            Source.Dispose();
        }

        public IEnumerator<TTarget> GetEnumerator()
        {
            return ((IEnumerable<T1>)Source).SelectMany(Func, Selector).GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}
