﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Text;

namespace NMF.Transformations.Linq
{
    /// <summary>
    /// Represents a growable wrapper for a given collection
    /// </summary>
    /// <typeparam name="T">The type of the arguments</typeparam>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1710:IdentifiersShouldHaveCorrectSuffix"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Growable")]
    public class GrowableSource<T> : IGrowableSource<T>, IDisposable
    {
        /// <summary>
        /// Gets the inner collection that is used to persist changes
        /// </summary>
        protected ICollection<T> Inner { get; private set; }

        private bool manualNotification;

        /// <summary>
        /// Creates a new growable collection wrapper for the given collection
        /// </summary>
        /// <param name="baseCollection">The collection that should be used as a basis</param>
        public GrowableSource(ICollection<T> baseCollection)
        {
            if (baseCollection == null) baseCollection = new List<T>();

            Inner = baseCollection;

            var notifier = baseCollection as INotifyCollectionChanged;

            if (notifier != null)
            {
                notifier.CollectionChanged += notifier_CollectionChanged;
                manualNotification = false;
            }
            else
            {
                manualNotification = true;
            }
        }

        /// <summary>
        /// Creates a new growable collection
        /// </summary>
        public GrowableSource() : this(null) { }

        void notifier_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Add)
            {
                foreach (T item in e.NewItems)
                {
                    NextItem(item);
                }
            }
        }

        /// <summary>
        /// Adds the given item to the collection
        /// </summary>
        /// <param name="item">The item that should be added to the collection</param>
        public void Add(T item)
        {
            Inner.Add(item);
            if (manualNotification) NextItem(item);
        }

        /// <summary>
        /// An event that is raised when a new item is added to the growable collection
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1710:IdentifiersShouldHaveCorrectSuffix")]
        public event Action<T> NextItem = delegate { };

        /// <summary>
        /// Gets an enumerator that enumerates the current collection
        /// </summary>
        /// <returns>An enumerator instance</returns>
        public IEnumerator<T> GetEnumerator()
        {
            return Inner.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        /// <summary>
        /// Disposes the current object
        /// </summary>
        /// <remarks>This simply unsets the notifier</remarks>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes the current object internally
        /// </summary>
        /// <param name="disposing">A value indicating whether the current call was made through a manually Dispose call</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                var notifier = Inner as INotifyCollectionChanged;
                if (notifier != null)
                {
                    notifier.CollectionChanged -= notifier_CollectionChanged;
                }
            }
        }
    }
}
